import datetime
import warnings
from functools import partial
from typing import Annotated, Optional, Union

from pydantic import BaseModel, ConfigDict, Field, model_validator

from docling.datamodel.base_models import DocumentStream

from docling_jobkit.datamodel.chunking import (
    ChunkingExportOptions,
    ChunkingOptionType,
)
from docling_jobkit.datamodel.convert import ConvertDocumentsOptions
from docling_jobkit.datamodel.http_inputs import FileSource, HttpSource
from docling_jobkit.datamodel.s3_coords import S3Coordinates
from docling_jobkit.datamodel.task_meta import TaskProcessingMeta, TaskStatus, TaskType
from docling_jobkit.datamodel.task_targets import InBodyTarget, TaskTarget

TaskSource = Union[HttpSource, FileSource, DocumentStream, S3Coordinates]


class Task(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    task_id: str
    task_type: TaskType = TaskType.CONVERT
    task_status: TaskStatus = TaskStatus.PENDING
    sources: list[TaskSource] = []
    target: TaskTarget = InBodyTarget()
    options: Annotated[
        Optional[ConvertDocumentsOptions],
        Field(
            description="Deprecated, use conversion_options instead.",
            deprecated="Use conversion_options instead.",
            exclude=True,
        ),
    ] = None
    convert_options: Optional[ConvertDocumentsOptions] = None
    chunking_options: Annotated[
        Optional[ChunkingOptionType],
        Field(discriminator="chunker"),
    ] = None
    chunking_export_options: ChunkingExportOptions = ChunkingExportOptions()
    # scratch_dir: Optional[Path] = None
    processing_meta: Optional[TaskProcessingMeta] = None
    created_at: datetime.datetime = Field(
        default_factory=partial(datetime.datetime.now, datetime.timezone.utc)
    )
    started_at: Optional[datetime.datetime] = None
    finished_at: Optional[datetime.datetime] = None
    last_update_at: datetime.datetime = Field(
        default_factory=partial(datetime.datetime.now, datetime.timezone.utc)
    )

    @model_validator(mode="before")
    def handle_deprecated_options(cls, values):
        # Warn users if they pass the deprecated field
        if "options" in values and "conversion_options" not in values:
            warnings.warn(
                "'options' is deprecated and will be removed in a future version. "
                "Use 'conversion_options' instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            values["conversion_options"] = values["options"]
        return values

    def set_status(self, status: TaskStatus):
        now = datetime.datetime.now(datetime.timezone.utc)
        if status == TaskStatus.STARTED and self.started_at is None:
            self.started_at = now
        if (
            status in [TaskStatus.SUCCESS, TaskStatus.FAILURE]
            and self.finished_at is None
        ):
            self.finished_at = now

        self.last_update_at = now
        self.task_status = status

    def is_completed(self) -> bool:
        if self.task_status in [TaskStatus.SUCCESS, TaskStatus.FAILURE]:
            return True
        return False



import logging
import os
import shutil
import time
from collections.abc import Iterable
from pathlib import Path

import httpx

from docling.datamodel.base_models import OutputFormat
from docling.datamodel.document import ConversionResult, ConversionStatus
from docling_core.types.doc import ImageRefMode

from docling_jobkit.datamodel.result import (
    DoclingTaskResult,
    ExportDocumentResponse,
    ExportResult,
    RemoteTargetResult,
    ResultType,
    ZipArchiveResult,
)
from docling_jobkit.datamodel.task import Task
from docling_jobkit.datamodel.task_targets import InBodyTarget, PutTarget

_log = logging.getLogger(__name__)


def _export_document_as_content(
    conv_res: ConversionResult,
    export_json: bool,
    export_html: bool,
    export_md: bool,
    export_txt: bool,
    export_doctags: bool,
    image_mode: ImageRefMode,
    md_page_break_placeholder: str,
) -> ExportDocumentResponse:
    document = ExportDocumentResponse(filename=conv_res.input.file.name)

    if conv_res.status == ConversionStatus.SUCCESS:
        new_doc = conv_res.document._make_copy_with_refmode(
            Path(), image_mode, page_no=None
        )

        # Create the different formats
        if export_json:
            document.json_content = new_doc
        if export_html:
            document.html_content = new_doc.export_to_html(image_mode=image_mode)
        if export_txt:
            document.text_content = new_doc.export_to_markdown(
                strict_text=True,
                image_mode=image_mode,
            )
        if export_md:
            document.md_content = new_doc.export_to_markdown(
                image_mode=image_mode,
                page_break_placeholder=md_page_break_placeholder or None,
            )
        if export_doctags:
            document.doctags_content = new_doc.export_to_doctags()

    return document


def _export_documents_as_files(
    conv_results: Iterable[ConversionResult],
    output_dir: Path,
    export_json: bool,
    export_html: bool,
    export_md: bool,
    export_txt: bool,
    export_doctags: bool,
    image_export_mode: ImageRefMode,
    md_page_break_placeholder: str,
):
    success_count = 0
    failure_count = 0

    artifacts_dir = Path("artifacts/")  # will be relative to the fname

    for conv_res in conv_results:
        if conv_res.status == ConversionStatus.SUCCESS:
            success_count += 1
            doc_filename = conv_res.input.file.stem

            # Export JSON format:
            if export_json:
                fname = output_dir / f"{doc_filename}.json"
                _log.info(f"writing JSON output to {fname}")
                conv_res.document.save_as_json(
                    filename=fname,
                    image_mode=image_export_mode,
                    artifacts_dir=artifacts_dir,
                )

            # Export HTML format:
            if export_html:
                fname = output_dir / f"{doc_filename}.html"
                _log.info(f"writing HTML output to {fname}")
                conv_res.document.save_as_html(
                    filename=fname,
                    image_mode=image_export_mode,
                    artifacts_dir=artifacts_dir,
                )

            # Export Text format:
            if export_txt:
                fname = output_dir / f"{doc_filename}.txt"
                _log.info(f"writing TXT output to {fname}")
                conv_res.document.save_as_markdown(
                    filename=fname,
                    strict_text=True,
                    image_mode=ImageRefMode.PLACEHOLDER,
                )

            # Export Markdown format:
            if export_md:
                fname = output_dir / f"{doc_filename}.md"
                _log.info(f"writing Markdown output to {fname}")
                conv_res.document.save_as_markdown(
                    filename=fname,
                    artifacts_dir=artifacts_dir,
                    image_mode=image_export_mode,
                    page_break_placeholder=md_page_break_placeholder or None,
                )

            # Export Document Tags format:
            if export_doctags:
                fname = output_dir / f"{doc_filename}.doctags"
                _log.info(f"writing Doc Tags output to {fname}")
                conv_res.document.save_as_doctags(filename=fname)

        else:
            _log.warning(f"Document {conv_res.input.file} failed to convert.")
            failure_count += 1

    conv_result = (
        ConversionStatus.SUCCESS if failure_count == 0 else ConversionStatus.FAILURE
    )

    _log.info(
        f"Processed {success_count + failure_count} docs, "
        f"of which {failure_count} failed"
    )
    return success_count, failure_count, conv_result


def process_export_results(
    task: Task,
    conv_results: Iterable[ConversionResult],
    work_dir: Path,
) -> DoclingTaskResult:
    conversion_options = task.convert_options
    if conversion_options is None:
        raise RuntimeError("process_export_results called without task.convert_options")

    # Let's start by processing the documents
    start_time = time.monotonic()

    # Convert the iterator to a list to count the number of results and get timings
    # As it's an iterator (lazy evaluation), it will also start the conversion
    conv_results = list(conv_results)

    processing_time = time.monotonic() - start_time

    _log.info(f"Processed {len(conv_results)} docs in {processing_time:.2f} seconds.")

    if len(conv_results) == 0:
        raise RuntimeError("No documents were generated by Docling.")

    # We have some results, let's prepare the response
    task_result: ResultType
    num_succeeded = 0
    num_failed = 0

    # Booleans to know what to export
    export_json = OutputFormat.JSON in conversion_options.to_formats
    export_html = OutputFormat.HTML in conversion_options.to_formats
    export_md = OutputFormat.MARKDOWN in conversion_options.to_formats
    export_txt = OutputFormat.TEXT in conversion_options.to_formats
    export_doctags = OutputFormat.DOCTAGS in conversion_options.to_formats

    # Only 1 document was processed, and we are not returning it as a file
    if len(conv_results) == 1 and isinstance(task.target, InBodyTarget):
        conv_res = conv_results[0]

        content = _export_document_as_content(
            conv_res,
            export_json=export_json,
            export_html=export_html,
            export_md=export_md,
            export_txt=export_txt,
            export_doctags=export_doctags,
            image_mode=conversion_options.image_export_mode,
            md_page_break_placeholder=conversion_options.md_page_break_placeholder,
        )
        task_result = ExportResult(
            content=content,
            status=conv_res.status,
            # processing_time=processing_time,
            timings=conv_res.timings,
            errors=conv_res.errors,
        )

        num_succeeded = 1 if conv_res.status == ConversionStatus.SUCCESS else 0
        num_failed = 1 if conv_res.status != ConversionStatus.SUCCESS else 0

    # Multiple documents were processed, or we are forced returning as a file
    else:
        # Temporary directory to store the outputs
        output_dir = work_dir / "output"
        output_dir.mkdir(parents=True, exist_ok=True)

        # Export the documents
        num_succeeded, num_failed, conv_result = _export_documents_as_files(
            conv_results=conv_results,
            output_dir=output_dir,
            export_json=export_json,
            export_html=export_html,
            export_md=export_md,
            export_txt=export_txt,
            export_doctags=export_doctags,
            image_export_mode=conversion_options.image_export_mode,
            md_page_break_placeholder=conversion_options.md_page_break_placeholder,
        )

        files = os.listdir(output_dir)
        if len(files) == 0:
            raise RuntimeError("No documents were exported.")

        file_path = work_dir / "converted_docs.zip"
        shutil.make_archive(
            base_name=str(file_path.with_suffix("")),
            format="zip",
            root_dir=output_dir,
        )

        if isinstance(task.target, PutTarget):
            try:
                with file_path.open("rb") as file_data:
                    r = httpx.put(str(task.target.url), files={"file": file_data})
                    r.raise_for_status()
                task_result = RemoteTargetResult()
            except Exception as exc:
                _log.error("An error occour while uploading zip to s3", exc_info=exc)
                raise RuntimeError(
                    "An error occour while uploading zip to the target url."
                )

        else:
            task_result = ZipArchiveResult(content=file_path.read_bytes())

    return DoclingTaskResult(
        result=task_result,
        processing_time=processing_time,
        num_succeeded=num_succeeded,
        num_failed=num_failed,
        num_converted=len(conv_results),
    )



from pathlib import Path
from typing import Annotated, Literal

from pydantic import BaseModel, Field

from docling_jobkit.datamodel.google_drive_coords import GoogleDriveCoordinates
from docling_jobkit.datamodel.http_inputs import FileSource, HttpSource
from docling_jobkit.datamodel.s3_coords import S3Coordinates


class TaskFileSource(FileSource):
    kind: Literal["file"] = "file"


class TaskHttpSource(HttpSource):
    kind: Literal["http"] = "http"


class TaskS3Source(S3Coordinates):
    kind: Literal["s3"] = "s3"


class TaskGoogleDriveSource(GoogleDriveCoordinates):
    kind: Literal["google_drive"] = "google_drive"


class TaskLocalPathSource(BaseModel):
    kind: Literal["local_path"] = "local_path"

    path: Annotated[
        Path,
        Field(
            description=(
                "Local filesystem path to a file or directory. "
                "For files, the single file will be processed. "
                "For directories, files will be discovered based on the pattern and recursive settings. "
                "Required."
            ),
            examples=[
                "/path/to/document.pdf",
                "/path/to/documents/",
                "./data/input/",
            ],
        ),
    ]

    pattern: Annotated[
        str,
        Field(
            description=(
                "Glob pattern for matching files within a directory. "
                "Supports standard glob syntax (e.g., '*.pdf', '**/*.docx'). "
                "Only applicable when path is a directory. "
                "Optional, defaults to '*' (all files)."
            ),
            examples=[
                "*.pdf",
                "*.{pdf,docx}",
                "**/*.pdf",
                "report_*.pdf",
            ],
        ),
    ] = "*"

    recursive: Annotated[
        bool,
        Field(
            description=(
                "If True, recursively traverse subdirectories when path is a directory. "
                "If False, only process files in the immediate directory. "
                "Optional, defaults to True."
            ),
        ),
    ] = True


TaskSource = Annotated[
    TaskFileSource
    | TaskHttpSource
    | TaskS3Source
    | TaskGoogleDriveSource
    | TaskLocalPathSource,
    Field(discriminator="kind"),
]



from io import BytesIO
from typing import Iterator

from docling.datamodel.base_models import DocumentStream

from docling_jobkit.connectors.google_drive_helper import GoogleDriveFileIdentifier
from docling_jobkit.connectors.source_processor import BaseSourceProcessor
from docling_jobkit.datamodel.google_drive_coords import GoogleDriveCoordinates


class GoogleDriveSourceProcessor(BaseSourceProcessor[GoogleDriveFileIdentifier]):
    def __init__(self, coords: GoogleDriveCoordinates):
        super().__init__()
        self._coords = coords

    def _initialize(self):
        from docling_jobkit.connectors.google_drive_helper import get_service

        self._service = get_service(self._coords)

    def _finalize(self):
        return

    def _fetch_documents(self) -> Iterator[DocumentStream]:
        from docling_jobkit.connectors.google_drive_helper import (
            download_file,
            get_source_files_infos,
        )

        files_infos = get_source_files_infos(
            service=self._service,
            coords=self._coords,
        )

        # download and yield one document at the time
        for file_info in files_infos:
            buffer = BytesIO()
            download_file(
                service=self._service,
                file_info=file_info,
                file_stream=buffer,
            )
            buffer.seek(0)

            yield DocumentStream(
                name=file_info["name"],
                stream=buffer,
            )

    def _list_document_ids(self) -> Iterator[GoogleDriveFileIdentifier]:
        from docling_jobkit.connectors.google_drive_helper import get_source_files_infos

        for info in get_source_files_infos(self._service, self._coords):
            yield GoogleDriveFileIdentifier(
                id=info["id"],
                name=info["name"],
                mimeType=info["mimeType"],
                path=info["path"],
            )

    def _fetch_document_by_id(self, info: GoogleDriveFileIdentifier) -> DocumentStream:
        from docling_jobkit.connectors.google_drive_helper import download_file

        buffer = BytesIO()

        download_file(
            service=self._service,
            file_info=info,
            file_stream=buffer,
        )
        buffer.seek(0)

        return DocumentStream(
            name=info["name"],
            stream=buffer,
        )



# Define the input options for the API
from typing import Annotated, Any, Optional

from pydantic import AnyUrl, BaseModel, Field, PositiveInt, model_validator
from typing_extensions import Self

from docling.datamodel import vlm_model_specs
from docling.datamodel.base_models import InputFormat, OutputFormat
from docling.datamodel.pipeline_options import (
    PdfBackend,
    PictureDescriptionBaseOptions,
    ProcessingPipeline,
    TableFormerMode,
    TableStructureOptions,
)
from docling.datamodel.pipeline_options_vlm_model import (
    InferenceFramework,
    InlineVlmOptions,
    ResponseFormat,
    TransformersModelType,
)
from docling.datamodel.settings import (
    DEFAULT_PAGE_RANGE,
    PageRange,
)
from docling_core.types.doc import ImageRefMode


class PictureDescriptionLocal(BaseModel):
    repo_id: Annotated[
        str,
        Field(
            description="Repository id from the Hugging Face Hub.",
            examples=[
                "HuggingFaceTB/SmolVLM-256M-Instruct",
                "ibm-granite/granite-vision-3.3-2b",
            ],
        ),
    ]
    prompt: Annotated[
        str,
        Field(
            description="Prompt used when calling the vision-language model.",
            examples=[
                "Describe this image in a few sentences.",
                "This is a figure from a document. Provide a detailed description of it.",
            ],
        ),
    ] = "Describe this image in a few sentences."
    generation_config: Annotated[
        dict[str, Any],
        Field(
            description="Config from https://huggingface.co/docs/transformers/en/main_classes/text_generation#transformers.GenerationConfig",
            examples=[{"max_new_tokens": 200, "do_sample": False}],
        ),
    ] = {"max_new_tokens": 200, "do_sample": False}


class PictureDescriptionApi(BaseModel):
    url: Annotated[
        AnyUrl,
        Field(
            description="Endpoint which accepts openai-api compatible requests.",
            examples=[
                AnyUrl(
                    "http://localhost:8000/v1/chat/completions"
                ),  # example of a local vllm api
                AnyUrl(
                    "http://localhost:1234/v1/chat/completions"
                ),  # example of lm studio
                AnyUrl(
                    "http://localhost:11434/v1/chat/completions"
                ),  # example of ollama
            ],
        ),
    ]
    headers: Annotated[
        dict[str, str],
        Field(
            description="Headers used for calling the API endpoint. For example, it could include authentication headers."
        ),
    ] = {}
    params: Annotated[
        dict[str, Any],
        Field(
            description="Model parameters.",
            examples=[
                {  # on vllm
                    "model": "HuggingFaceTB/SmolVLM-256M-Instruct",
                    "max_completion_tokens": 200,
                },
                {  # on vllm
                    "model": "ibm-granite/granite-vision-3.3-2b",
                    "max_completion_tokens": 200,
                },
                {  # on ollama
                    "model": "granite3.2-vision:2b"
                },
            ],
        ),
    ] = {}
    timeout: Annotated[float, Field(description="Timeout for the API request.")] = 20
    concurrency: Annotated[
        PositiveInt,
        Field(
            description="Maximum number of concurrent requests to the API.",
            examples=[1],
        ),
    ] = 1
    prompt: Annotated[
        str,
        Field(
            description="Prompt used when calling the vision-language model.",
            examples=[
                "Describe this image in a few sentences.",
                "This is a figures from a document. Provide a detailed description of it.",
            ],
        ),
    ] = "Describe this image in a few sentences."


class VlmModelLocal(BaseModel):
    repo_id: Annotated[
        str,
        Field(
            description="Repository id from the Hugging Face Hub.",
        ),
    ]
    prompt: Annotated[
        str,
        Field(
            description="Prompt used when calling the vision-language model.",
            examples=[
                "Convert this page to docling.",
                "Convert this page to markdown. Do not miss any text and only output the bare markdown!",
            ],
        ),
    ] = "Convert this page to docling."
    scale: Annotated[float, Field(description="Scale factor of the images used.")] = 2.0
    response_format: Annotated[
        ResponseFormat, Field(description="Type of response generated by the model.")
    ]
    inference_framework: Annotated[
        InferenceFramework, Field(description="Inference framework to use.")
    ]
    transformers_model_type: Annotated[
        TransformersModelType,
        Field(description="Type of transformers auto-model to use."),
    ] = TransformersModelType.AUTOMODEL
    extra_generation_config: Annotated[
        dict[str, Any],
        Field(
            description="Config from https://huggingface.co/docs/transformers/en/main_classes/text_generation#transformers.GenerationConfig",
            examples=[{"max_new_tokens": 800, "do_sample": False}],
        ),
    ] = {"max_new_tokens": 800, "do_sample": False}
    temperature: Annotated[
        float,
        Field(
            description="Temperature parameter controlling the reproducibility of the result.",
            examples=[0.0, 1.0],
        ),
    ] = 0.0

    @staticmethod
    def from_docling(options: InlineVlmOptions):
        return VlmModelLocal.model_validate(options.model_dump())


class VlmModelApi(BaseModel):
    url: Annotated[
        AnyUrl,
        Field(
            description="Endpoint which accepts openai-api compatible requests.",
            examples=[
                AnyUrl(
                    "http://localhost:8000/v1/chat/completions"
                ),  # example of a local vllm api
                AnyUrl(
                    "http://localhost:1234/v1/chat/completions"
                ),  # example of lm studio
            ],
        ),
    ]
    headers: Annotated[
        dict[str, str],
        Field(
            description="Headers used for calling the API endpoint. For example, it could include authentication headers."
        ),
    ] = {}
    params: Annotated[
        dict[str, Any],
        Field(
            description="Model parameters.",
            examples=[
                {  # on vllm
                    "model": "ibm-granite/granite-docling-258M",
                    "max_completion_tokens": 800,
                },
                {  # on vllm
                    "model": "ibm-granite/granite-vision-3.3-2b",
                    "max_completion_tokens": 800,
                },
            ],
        ),
    ] = {}
    timeout: Annotated[float, Field(description="Timeout for the API request.")] = 60
    concurrency: Annotated[
        PositiveInt,
        Field(
            description="Maximum number of concurrent requests to the API.",
            examples=[1],
        ),
    ] = 1
    prompt: Annotated[
        str,
        Field(
            description="Prompt used when calling the vision-language model.",
            examples=[
                "Convert this page to docling.",
                "Convert this page to markdown. Do not miss any text and only output the bare markdown!",
            ],
        ),
    ] = "Convert this page to docling."
    scale: Annotated[float, Field(description="Scale factor of the images used.")] = 2.0
    response_format: Annotated[
        ResponseFormat, Field(description="Type of response generated by the model.")
    ]
    temperature: Annotated[
        float,
        Field(
            description="Temperature parameter controlling the reproducibility of the result.",
            examples=[0.0, 1.0],
        ),
    ] = 0.0


class ConvertDocumentsOptions(BaseModel):
    from_formats: Annotated[
        list[InputFormat],
        Field(
            description=(
                "Input format(s) to convert from. String or list of strings. "
                f"Allowed values: {', '.join([v.value for v in InputFormat])}. "
                "Optional, defaults to all formats."
            ),
            examples=[[v.value for v in InputFormat]],
        ),
    ] = list(InputFormat)

    to_formats: Annotated[
        list[OutputFormat],
        Field(
            description=(
                "Output format(s) to convert to. String or list of strings. "
                f"Allowed values: {', '.join([v.value for v in OutputFormat])}. "
                "Optional, defaults to Markdown."
            ),
            examples=[
                [OutputFormat.MARKDOWN],
                [OutputFormat.MARKDOWN, OutputFormat.JSON],
                [v.value for v in OutputFormat],
            ],
        ),
    ] = [OutputFormat.MARKDOWN]

    image_export_mode: Annotated[
        ImageRefMode,
        Field(
            description=(
                "Image export mode for the document (in case of JSON,"
                " Markdown or HTML). "
                f"Allowed values: {', '.join([v.value for v in ImageRefMode])}. "
                "Optional, defaults to Embedded."
            ),
            examples=[ImageRefMode.EMBEDDED.value],
            # pattern="embedded|placeholder|referenced",
        ),
    ] = ImageRefMode.EMBEDDED

    do_ocr: Annotated[
        bool,
        Field(
            description=(
                "If enabled, the bitmap content will be processed using OCR. "
                "Boolean. Optional, defaults to true"
            ),
            # examples=[True],
        ),
    ] = True

    force_ocr: Annotated[
        bool,
        Field(
            description=(
                "If enabled, replace existing text with OCR-generated "
                "text over content. Boolean. Optional, defaults to false."
            ),
            # examples=[False],
        ),
    ] = False

    ocr_engine: Annotated[
        str,
        Field(
            description=("The OCR engine to use. String. "),
        ),
    ] = "auto"

    ocr_lang: Annotated[
        Optional[list[str]],
        Field(
            description=(
                "List of languages used by the OCR engine. "
                "Note that each OCR engine has "
                "different values for the language names. String or list of strings. "
                "Optional, defaults to empty."
            ),
            examples=[["fr", "de", "es", "en"]],
        ),
    ] = None

    pdf_backend: Annotated[
        PdfBackend,
        Field(
            description=(
                "The PDF backend to use. String. "
                f"Allowed values: {', '.join([v.value for v in PdfBackend])}. "
                f"Optional, defaults to {PdfBackend.DLPARSE_V4.value}."
            ),
            examples=[PdfBackend.DLPARSE_V4],
        ),
    ] = PdfBackend.DLPARSE_V4

    table_mode: Annotated[
        TableFormerMode,
        Field(
            description=(
                "Mode to use for table structure, String. "
                f"Allowed values: {', '.join([v.value for v in TableFormerMode])}. "
                "Optional, defaults to accurate."
            ),
            examples=[TableStructureOptions().mode],
            # pattern="fast|accurate",
        ),
    ] = TableStructureOptions().mode

    table_cell_matching: Annotated[
        bool,
        Field(
            description="If true, matches table cells predictions back to PDF cells. Can break table output if PDF cells are merged across table columns. If false, let table structure model define the text cells, ignore PDF cells.",
            examples=[TableStructureOptions().do_cell_matching],
        ),
    ] = TableStructureOptions().do_cell_matching

    pipeline: Annotated[
        ProcessingPipeline,
        Field(description="Choose the pipeline to process PDF or image files."),
    ] = ProcessingPipeline.STANDARD

    page_range: Annotated[
        PageRange,
        Field(
            description="Only convert a range of pages. The page number starts at 1.",
            examples=[DEFAULT_PAGE_RANGE, (1, 4)],
        ),
    ] = DEFAULT_PAGE_RANGE

    document_timeout: Annotated[
        Optional[float],
        Field(
            description="The timeout for processing each document, in seconds.",
        ),
    ] = None

    abort_on_error: Annotated[
        bool,
        Field(
            description=(
                "Abort on error if enabled. Boolean. Optional, defaults to false."
            ),
            # examples=[False],
        ),
    ] = False

    do_table_structure: Annotated[
        bool,
        Field(
            description=(
                "If enabled, the table structure will be extracted. "
                "Boolean. Optional, defaults to true."
            ),
            examples=[True],
        ),
    ] = True

    include_images: Annotated[
        bool,
        Field(
            description=(
                "If enabled, images will be extracted from the document. "
                "Boolean. Optional, defaults to true."
            ),
            examples=[True],
        ),
    ] = True

    images_scale: Annotated[
        float,
        Field(
            description="Scale factor for images. Float. Optional, defaults to 2.0.",
            examples=[2.0],
        ),
    ] = 2.0

    md_page_break_placeholder: Annotated[
        str,
        Field(
            description="Add this placeholder between pages in the markdown output.",
            examples=["<!-- page-break -->", ""],
        ),
    ] = ""

    do_code_enrichment: Annotated[
        bool,
        Field(
            description=(
                "If enabled, perform OCR code enrichment. "
                "Boolean. Optional, defaults to false."
            ),
            examples=[False],
        ),
    ] = False

    do_formula_enrichment: Annotated[
        bool,
        Field(
            description=(
                "If enabled, perform formula OCR, return LaTeX code. "
                "Boolean. Optional, defaults to false."
            ),
            examples=[False],
        ),
    ] = False

    do_picture_classification: Annotated[
        bool,
        Field(
            description=(
                "If enabled, classify pictures in documents. "
                "Boolean. Optional, defaults to false."
            ),
            examples=[False],
        ),
    ] = False

    do_picture_description: Annotated[
        bool,
        Field(
            description=(
                "If enabled, describe pictures in documents. "
                "Boolean. Optional, defaults to false."
            ),
            examples=[False],
        ),
    ] = False

    picture_description_area_threshold: Annotated[
        float,
        Field(
            description="Minimum percentage of the area for a picture to be processed with the models.",
            examples=[PictureDescriptionBaseOptions().picture_area_threshold],
        ),
    ] = PictureDescriptionBaseOptions().picture_area_threshold

    picture_description_local: Annotated[
        Optional[PictureDescriptionLocal],
        Field(
            description="Options for running a local vision-language model in the picture description. The parameters refer to a model hosted on Hugging Face. This parameter is mutually exclusive with picture_description_api.",
            examples=[
                PictureDescriptionLocal(repo_id="ibm-granite/granite-vision-3.2-2b"),
                PictureDescriptionLocal(repo_id="HuggingFaceTB/SmolVLM-256M-Instruct"),
            ],
        ),
    ] = None

    picture_description_api: Annotated[
        Optional[PictureDescriptionApi],
        Field(
            description="API details for using a vision-language model in the picture description. This parameter is mutually exclusive with picture_description_local.",
            examples=[
                PictureDescriptionApi(
                    url="http://localhost:1234/v1/chat/completions",
                    params={"model": "granite3.2-vision:2b"},
                ),
                PictureDescriptionApi(
                    url="http://localhost:11434/v1/chat/completions",
                    params={"model": "granite3.2-vision:2b"},
                ),
            ],
        ),
    ] = None

    vlm_pipeline_model: Annotated[
        Optional[vlm_model_specs.VlmModelType],
        Field(
            description="Preset of local and API models for the vlm pipeline. This parameter is mutually exclusive with vlm_pipeline_model_local and vlm_pipeline_model_api. Use the other options for more parameters.",
            examples=[vlm_model_specs.VlmModelType.GRANITEDOCLING],
        ),
    ] = None
    vlm_pipeline_model_local: Annotated[
        Optional[VlmModelLocal],
        Field(
            description="Options for running a local vision-language model for the vlm pipeline. The parameters refer to a model hosted on Hugging Face. This parameter is mutually exclusive with vlm_pipeline_model_api and vlm_pipeline_model.",
            examples=[
                VlmModelLocal.from_docling(vlm_model_specs.GRANITEDOCLING_TRANSFORMERS),
                VlmModelLocal.from_docling(vlm_model_specs.GRANITEDOCLING_MLX),
                VlmModelLocal.from_docling(vlm_model_specs.GRANITE_VISION_TRANSFORMERS),
            ],
        ),
    ] = None

    vlm_pipeline_model_api: Annotated[
        Optional[VlmModelApi],
        Field(
            description="API details for using a vision-language model for the vlm pipeline. This parameter is mutually exclusive with vlm_pipeline_model_local and vlm_pipeline_model.",
            examples=[
                VlmModelApi(
                    url="http://localhost:1234/v1/chat/completions",
                    params={"model": "ibm-granite/granite-docling-258M-mlx"},
                    response_format=ResponseFormat.DOCTAGS,
                    prompt="Convert this page to docling.",
                )
            ],
        ),
    ] = None

    @model_validator(mode="after")
    def picture_description_exclusivity(self) -> Self:
        # Validate picture description options
        if (
            self.picture_description_local is not None
            and self.picture_description_api is not None
        ):
            raise ValueError(
                "The parameters picture_description_local and picture_description_api are mutually exclusive, only one of them can be set."
            )

        return self

    @model_validator(mode="after")
    def vlm_model_exclusivity(self) -> Self:
        # Validate vlm model options
        num_not_nan = sum(
            (
                opt is not None
                for opt in (
                    self.vlm_pipeline_model,
                    self.vlm_pipeline_model_local,
                    self.vlm_pipeline_model_api,
                )
            )
        )
        if num_not_nan > 1:
            raise ValueError(
                "The parameters vlm_pipeline_model, vlm_pipeline_model_local and vlm_pipeline_model_api are mutually exclusive, only one of them can be set."
            )

        return self



from typing import Iterator, TypedDict

from docling_core.types.io import DocumentStream

from docling_jobkit.connectors.source_processor import BaseSourceProcessor
from docling_jobkit.datamodel.http_inputs import FileSource, HttpSource


class HttpFileIdentifier(TypedDict):
    source: HttpSource | FileSource
    index: int


class HttpSourceProcessor(BaseSourceProcessor[HttpFileIdentifier]):
    def __init__(self, source: HttpSource | FileSource):
        super().__init__()
        self._source = source

    def _initialize(self):
        pass

    def _finalize(self):
        pass

    def _list_document_ids(self) -> Iterator[HttpFileIdentifier]:
        """Yield a single identifier for the HTTP/File source."""
        yield HttpFileIdentifier(source=self._source, index=0)

    def _fetch_document_by_id(self, identifier: HttpFileIdentifier) -> DocumentStream:
        """Fetch document from the identifier."""
        source = identifier["source"]
        if isinstance(source, FileSource):
            return source.to_document_stream()
        elif isinstance(source, HttpSource):
            # TODO: fetch, e.g. using the helpers in docling-core
            raise NotImplementedError("HttpSource fetching is not yet implemented")
        else:
            raise ValueError(f"Unsupported source type: {type(source)}")

    def _fetch_documents(self) -> Iterator[DocumentStream]:
        if isinstance(self._source, FileSource):
            yield self._source.to_document_stream()
        elif isinstance(self._source, HttpSource):
            # TODO: fetch, e.g. using the helpers in docling-core
            raise NotImplementedError()



# ruff: noqa: E402

from typing import List, NamedTuple

from kfp import dsl, kubernetes
from kfp.dsl import Dataset, Input, Output


@dsl.component(
    packages_to_install=[
        "docling==2.50.0",
        "boto3~=1.35.36",
        "git+https://github.com/docling-project/docling-jobkit@main",
    ],
    base_image="ghcr.io/docling-project/docling-serve-cpu:v1.5.1",  # base docling-serve image with fixed permissions
)
def convert_payload(
    options: dict,
    source: dict,
    target: dict,
    batch_index: int,
    # source_keys: List[str],
    dataset: Input[Dataset],
) -> list:
    import json
    import logging
    from pathlib import Path

    from docling_jobkit.connectors.s3_helper import (
        generate_presign_url,
        get_s3_connection,
    )
    from docling_jobkit.connectors.s3_target_processor import S3TargetProcessor
    from docling_jobkit.convert.manager import (
        DoclingConverterManager,
        DoclingConverterManagerConfig,
    )
    from docling_jobkit.convert.results_processor import ResultsProcessor
    from docling_jobkit.datamodel.convert import ConvertDocumentsOptions
    from docling_jobkit.datamodel.s3_coords import S3Coordinates

    logging.basicConfig(level=logging.INFO)

    # validate coords
    s3_coords_source = S3Coordinates.model_validate(source)
    target_s3_coords = S3Coordinates.model_validate(target)

    s3_source_client, s3_source_resource = get_s3_connection(s3_coords_source)

    convert_options = ConvertDocumentsOptions.model_validate(options)

    config = DoclingConverterManagerConfig()
    # Points to the pvc mounted on the pod
    config.artifacts_path = Path("/modelcache")
    converter = DoclingConverterManager(config)

    with open(dataset.path) as f:
        batches = json.load(f)
    source_keys = batches[batch_index]

    presign_filtered_source_keys = [
        url
        for url in [
            generate_presign_url(s3_source_client, key, s3_coords_source.bucket)
            for key in source_keys
        ]
        if url is not None
    ]

    results = []
    with S3TargetProcessor(target_s3_coords) as target_processor:
        result_processor = ResultsProcessor(
            target_processor=target_processor,
            to_formats=[v.value for v in convert_options.to_formats],
            generate_page_images=convert_options.include_images,
            generate_picture_images=convert_options.include_images,
        )
        for item in result_processor.process_documents(
            converter.convert_documents(
                presign_filtered_source_keys, options=convert_options
            )
        ):
            results.append(item)
            logging.info("Convertion result: {}".format(item))

    return results


@dsl.component(
    packages_to_install=[
        "pydantic",
        "boto3~=1.35.36",
        "git+https://github.com/docling-project/docling-jobkit@main",
    ],
    base_image="ghcr.io/docling-project/docling-serve-cpu:v1.5.1",
)
def compute_batches(
    source: dict,
    target: dict,
    dataset: Output[Dataset],
    batch_size: int = 10,
) -> NamedTuple("outputs", [("batch_indices", List[int])]):  # type: ignore[valid-type]
    import json
    import logging
    from typing import NamedTuple

    from docling_jobkit.connectors.s3_helper import (
        check_target_has_source_converted,
        generate_batch_keys,
        get_s3_connection,
        get_source_files,
    )
    from docling_jobkit.datamodel.s3_coords import S3Coordinates

    logging.basicConfig(level=logging.INFO)

    # validate inputs
    s3_coords_source = S3Coordinates.model_validate(source)
    s3_target_coords = S3Coordinates.model_validate(target)

    s3_source_client, s3_source_resource = get_s3_connection(s3_coords_source)
    source_objects_list = get_source_files(
        s3_source_client, s3_source_resource, s3_coords_source
    )
    filtered_source_keys = check_target_has_source_converted(
        s3_target_coords, source_objects_list, s3_coords_source.key_prefix
    )
    batch_keys = generate_batch_keys(
        filtered_source_keys,
        batch_size=batch_size,
    )

    # store batches on s3 for debugging
    s3_target_client, s3_target_resource = get_s3_connection(s3_target_coords)
    try:
        s3_target_client.put_object(
            Body=json.dumps(batch_keys),
            Bucket=s3_target_coords.bucket,
            Key=f"{s3_target_coords.key_prefix}/buckets.json",
            ContentType="application/json",
        )
    except Exception as e:
        logging.error("Uploading batch.json to s3 failed: {}".format(e))

    with open(dataset.path, "w") as out_batches:
        json.dump(batch_keys, out_batches)

    batch_indices = list(range(len(batch_keys)))
    outputs = NamedTuple("outputs", [("batch_indices", List[int])])
    print(f"Batches to convert: {len(batch_keys)}")
    return outputs(batch_indices)


@dsl.pipeline
def inputs_s3in_s3out(
    convertion_options: dict = {
        "from_formats": [
            "docx",
            "pptx",
            "html",
            "image",
            "pdf",
            "asciidoc",
            "md",
            "xlsx",
            "xml_uspto",
            "xml_jats",
            "json_docling",
        ],
        "to_formats": ["md", "json", "html", "text", "doctags"],
        "image_export_mode": "placeholder",
        "do_ocr": True,
        "force_ocr": False,
        "ocr_engine": "easyocr",
        "ocr_lang": [],
        "pdf_backend": "dlparse_v2",
        "table_mode": "accurate",
        "abort_on_error": False,
        "return_as_file": False,
        "do_table_structure": True,
        "do_code_enrichment": False,
        "do_formula_enrichment": False,
        "do_picture_classification": False,
        "do_picture_description": False,
        "generate_picture_images": False,
        "include_images": True,
        "images_scale": 2,
    },
    source: dict = {
        "endpoint": "s3.eu-de.cloud-object-storage.appdomain.cloud",
        "access_key": "123454321",
        "secret_key": "secretsecret",
        "bucket": "source-bucket",
        "key_prefix": "my-docs",
        "verify_ssl": True,
    },
    target: dict = {
        "endpoint": "s3.eu-de.cloud-object-storage.appdomain.cloud",
        "access_key": "123454321",
        "secret_key": "secretsecret",
        "bucket": "target-bucket",
        "key_prefix": "my-docs",
        "verify_ssl": True,
    },
    batch_size: int = 20,
):
    import logging

    logging.basicConfig(level=logging.INFO)

    batches = compute_batches(source=source, target=target, batch_size=batch_size)
    # disable caching on batches as cached pre-signed urls might have already expired
    batches.set_caching_options(False)

    results = []
    with dsl.ParallelFor(batches.outputs["batch_indices"], parallelism=20) as subbatch:
        converter = convert_payload(
            options=convertion_options,
            source=source,
            target=target,
            dataset=batches.outputs["dataset"],
            batch_index=subbatch,
        )
        converter.set_memory_request("1G")
        converter.set_memory_limit("16G")
        converter.set_cpu_request("200m")
        converter.set_cpu_limit("1")

        # All models should be preloaded to pvc, multi access pvc can be used by all pods in parallel
        kubernetes.mount_pvc(
            converter,
            pvc_name="docling-model-cache-pvc-multi",
            mount_path="/modelcache",
        )

        # For enabling document conversion using GPU
        # currently unable to properly pass input parameters into pipeline, therefore node selector and tolerations are hardcoded

        converter.set_accelerator_type("nvidia.com/gpu")
        converter.set_accelerator_limit("1")

        # kubernetes.add_node_selector(
        #     task=converter,
        #     label_key="nvidia.com/gpu.product",
        #     label_value="NVIDIA-A10",
        # )

        kubernetes.add_toleration(
            task=converter,
            key="key1",
            operator="Equal",
            value="mcad",
            effect="NoSchedule",
        )

        results.append(converter.output)


### Compile pipeline into a yaml
from kfp import compiler

compiler.Compiler().compile(inputs_s3in_s3out, "docling-s3in-s3out.yaml")


### Start pipeline run programatically
# import kfp
# import os

# # TIP: you may need to authenticate with the KFP instance
# kfp_client = kfp.Client(
#     host = os.environ["KFP_FULL_URL"],
#     existing_token = os.environ["OPENSHIFT_TOKEN"],
#     verify_ssl = False,
# )

# kfp_client.create_run_from_pipeline_func(
#     inputs_s3in_s3out,
#     arguments=dict(
#         convertion_options = {
#             "from_formats": [
#                 "pdf",
#             ],
#             "to_formats": ["md", "json", "html", "text", "doctags"],
#             "image_export_mode": "placeholder",
#             "do_ocr": True,
#             "force_ocr": True,
#             "ocr_engine": "easyocr",
#         },
#         source = {
#             "endpoint": os.environ["S3_SOURCE_ENDPOINT"],
#             "access_key": os.environ["S3_SOURCE_ACCESS_KEY"],
#             "secret_key": os.environ["S3_SOURCE_SECRET_KEY"],
#             "bucket": os.environ["S3_SOURCE_BUCKET"],
#             "key_prefix": os.environ["S3_SOURCE_PREFIX"],
#             "verify_ssl": True,
#         },
#         target = {
#             "endpoint": os.environ["S3_TARGET_ENDPOINT"],
#             "access_key": os.environ["S3_TARGET_ACCESS_KEY"],
#             "secret_key": os.environ["S3_TARGET_SECRET_KEY"],
#             "bucket": os.environ["S3_TARGET_BUCKET"],
#             "key_prefix": os.environ["S3_TARGET_PREFIX"],
#             "verify_ssl": True,
#         },
#         batch_size = 100,
#     )
# )



from pathlib import Path
from typing import BinaryIO

from docling_jobkit.connectors.target_processor import BaseTargetProcessor
from docling_jobkit.datamodel.task_targets import LocalPathTarget


class LocalPathTargetProcessor(BaseTargetProcessor):
    def __init__(self, target: LocalPathTarget):
        super().__init__()
        self._target = target

    def _initialize(self) -> None:
        """
        Ensure the target directory exists.
        If path is a directory, create it. If it's a file path, create parent directories.
        """
        path = self._target.path

        # If path looks like a directory (ends with / or has no extension), treat as directory
        # Otherwise, create parent directories for the file
        if path.suffix == "" or str(path).endswith("/"):
            # Treat as directory
            path.mkdir(parents=True, exist_ok=True)
        else:
            # Treat as file - create parent directories
            path.parent.mkdir(parents=True, exist_ok=True)

    def _finalize(self) -> None:
        """No cleanup needed for local filesystem."""

    def upload_file(
        self,
        filename: str | Path,
        target_filename: str,
        content_type: str,
    ) -> None:
        """
        Copy a file from local filesystem to the target location.
        """
        source_path = Path(filename)
        target_path = self._get_target_path(target_filename)

        # Ensure parent directory exists
        target_path.parent.mkdir(parents=True, exist_ok=True)

        # Copy file content
        with open(source_path, "rb") as src:
            with open(target_path, "wb") as dst:
                dst.write(src.read())

    def upload_object(
        self,
        obj: str | bytes | BinaryIO,
        target_filename: str,
        content_type: str,
    ) -> None:
        """
        Write an in-memory object (bytes or file-like) to the target location.
        """
        target_path = self._get_target_path(target_filename)

        # Ensure parent directory exists
        target_path.parent.mkdir(parents=True, exist_ok=True)

        # Write content based on type
        if isinstance(obj, str):
            with open(target_path, "w", encoding="utf-8") as f:
                f.write(obj)
        elif isinstance(obj, (bytes, bytearray)):
            with open(target_path, "wb") as f:
                f.write(obj)
        else:
            # Assume it's a file-like object
            with open(target_path, "wb") as f:
                f.write(obj.read())

    def _get_target_path(self, target_filename: str) -> Path:
        """
        Determine the full target path based on the configured path.
        - If path is a directory, append target_filename
        - If path is a file, use it directly (ignore target_filename)
        """
        path = self._target.path

        # Check if path is intended to be a directory
        if path.is_dir() or path.suffix == "" or str(path).endswith("/"):
            # Treat as directory - append target_filename
            return path / target_filename
        else:
            # Treat as file - use the path directly
            return path



from io import BytesIO
from pathlib import Path
from typing import BinaryIO

from docling_jobkit.connectors.target_processor import BaseTargetProcessor
from docling_jobkit.datamodel.google_drive_coords import GoogleDriveCoordinates


class GoogleDriveTargetProcessor(BaseTargetProcessor):
    def __init__(self, coords: GoogleDriveCoordinates):
        super().__init__()
        self._coords = coords

    def _initialize(self):
        from docling_jobkit.connectors.google_drive_helper import get_service

        self._service = get_service(self._coords)

    def _finalize(self):
        return

    def upload_file(
        self,
        filename: str | Path,
        target_filename: str,
        content_type: str,
    ) -> None:
        """
        Upload a local file from disk to Google Drive.
        """
        from docling_jobkit.connectors.google_drive_helper import upload_file

        upload_file(
            service=self._service,
            filename=filename,
            target_filename=target_filename,
            content_type=content_type,
            coords=self._coords,
        )

    def upload_object(
        self,
        obj: str | bytes | BinaryIO,
        target_filename: str,
        content_type: str,
    ) -> None:
        """
        Upload an in-memory object (bytes or file-like) to Google Drive.
        """
        from docling_jobkit.connectors.google_drive_helper import upload_file

        if isinstance(obj, (bytes, bytearray)):
            body: BinaryIO = BytesIO(obj)
        elif isinstance(obj, str):
            body = BytesIO(obj.encode())
        else:
            body = obj  # assume it's a file-like object

        upload_file(
            service=self._service,
            file_stream=body,
            target_filename=target_filename,
            content_type=content_type,
            coords=self._coords,
        )



import asyncio
import gc
import logging
import tempfile
import uuid
import warnings
from pathlib import Path
from typing import Optional

from pydantic import BaseModel

from docling.datamodel.base_models import InputFormat

from docling_jobkit.convert.chunking import DocumentChunkerManager
from docling_jobkit.convert.manager import DoclingConverterManager
from docling_jobkit.datamodel.chunking import BaseChunkerOptions, ChunkingExportOptions
from docling_jobkit.datamodel.convert import ConvertDocumentsOptions
from docling_jobkit.datamodel.result import DoclingTaskResult
from docling_jobkit.datamodel.task import Task, TaskSource, TaskTarget
from docling_jobkit.datamodel.task_meta import TaskType
from docling_jobkit.orchestrators.base_orchestrator import (
    BaseOrchestrator,
)
from docling_jobkit.orchestrators.local.worker import AsyncLocalWorker

_log = logging.getLogger(__name__)


class LocalOrchestratorConfig(BaseModel):
    num_workers: int = 2
    shared_models: bool = False
    scratch_dir: Optional[Path] = None


class LocalOrchestrator(BaseOrchestrator):
    def __init__(
        self,
        config: LocalOrchestratorConfig,
        converter_manager: DoclingConverterManager,
    ):
        super().__init__()
        self.config = config
        self.task_queue: asyncio.Queue[str] = asyncio.Queue()
        self.queue_list: list[str] = []
        self.cm = converter_manager
        self.chunker_manager = DocumentChunkerManager()
        self.worker_cms: list[DoclingConverterManager] = []
        self._task_results: dict[str, DoclingTaskResult] = {}
        self.scratch_dir = self.config.scratch_dir or Path(
            tempfile.mkdtemp(prefix="docling_")
        )

    async def enqueue(
        self,
        sources: list[TaskSource],
        target: TaskTarget,
        task_type: TaskType = TaskType.CONVERT,
        options: ConvertDocumentsOptions | None = None,
        convert_options: ConvertDocumentsOptions | None = None,
        chunking_options: BaseChunkerOptions | None = None,
        chunking_export_options: ChunkingExportOptions | None = None,
    ) -> Task:
        if options is not None and convert_options is None:
            convert_options = options
            warnings.warn(
                "'options' is deprecated and will be removed in a future version. "
                "Use 'conversion_options' instead.",
                DeprecationWarning,
                stacklevel=2,
            )
        task_id = str(uuid.uuid4())
        chunking_export_options = chunking_export_options or ChunkingExportOptions()
        task = Task(
            task_id=task_id,
            task_type=task_type,
            sources=sources,
            convert_options=convert_options,
            chunking_options=chunking_options,
            chunking_export_options=chunking_export_options,
            target=target,
        )
        await self.init_task_tracking(task)

        self.queue_list.append(task_id)
        await self.task_queue.put(task_id)
        return task

    async def queue_size(self) -> int:
        return self.task_queue.qsize()

    async def get_queue_position(self, task_id: str) -> Optional[int]:
        return (
            self.queue_list.index(task_id) + 1 if task_id in self.queue_list else None
        )

    async def task_result(
        self,
        task_id: str,
    ) -> Optional[DoclingTaskResult]:
        if task_id not in self._task_results:
            return None
        return self._task_results[task_id]

    async def process_queue(self):
        # Create a pool of workers
        workers = []
        for i in range(self.config.num_workers):
            _log.debug(f"Starting worker {i}")
            w = AsyncLocalWorker(
                i,
                self,
                use_shared_manager=self.config.shared_models,
                scratch_dir=self.scratch_dir,
            )
            worker_task = asyncio.create_task(w.loop())
            workers.append(worker_task)

        # Wait for all workers to complete (they won't, as they run indefinitely)
        await asyncio.gather(*workers)
        _log.debug("All workers completed.")

    async def warm_up_caches(self):
        # Converter with default options
        pdf_format_option = self.cm.get_pdf_pipeline_opts(ConvertDocumentsOptions())
        converter = self.cm.get_converter(pdf_format_option)
        converter.initialize_pipeline(InputFormat.PDF)

    async def delete_task(self, task_id: str):
        _log.info(f"Deleting result of task {task_id=}")
        if task_id in self._task_results:
            del self._task_results[task_id]
        await super().delete_task(task_id)

    async def clear_converters(self):
        self.cm.clear_cache()
        self.chunker_manager.clear_cache()
        for cm in self.worker_cms:
            cm.clear_cache()
        gc.collect()

    async def check_connection(self):
        pass



from io import BytesIO
from typing import Iterator, TypedDict

from docling_core.types.io import DocumentStream

from docling_jobkit.connectors.s3_helper import get_s3_connection
from docling_jobkit.connectors.source_processor import BaseSourceProcessor
from docling_jobkit.datamodel.s3_coords import S3Coordinates


class S3FileIdentifier(TypedDict):
    key: str  # S3 object key
    size: int  # optional, include if available
    last_modified: str | None  # ISO timestamp, optional


class S3SourceProcessor(BaseSourceProcessor[S3FileIdentifier]):
    def __init__(self, coords: S3Coordinates):
        super().__init__()
        self._coords = coords

    def _initialize(self):
        self._client, self._resource = get_s3_connection(self._coords)

    def _finalize(self):
        self._client.close()

    def _list_document_ids(self) -> Iterator[S3FileIdentifier]:
        paginator = self._client.get_paginator("list_objects_v2")
        for page in paginator.paginate(
            Bucket=self._coords.bucket,
            Prefix=self._coords.key_prefix,
        ):
            for obj in page.get("Contents", []):
                last_modified = obj.get("LastModified", None)
                yield S3FileIdentifier(
                    key=obj["Key"],  # type: ignore[typeddict-item]  # Key is always present in S3 list_objects_v2 response
                    size=obj.get("Size", 0),
                    last_modified=last_modified.isoformat() if last_modified else None,
                )

    def _count_documents(self) -> int:
        total = 0
        paginator = self._client.get_paginator("list_objects_v2")
        for page in paginator.paginate(
            Bucket=self._coords.bucket,
            Prefix=self._coords.key_prefix,
        ):
            total += len(page.get("Contents", []))
        return total

    # ----------------- Document fetch -----------------

    def _fetch_document_by_id(self, identifier: S3FileIdentifier) -> DocumentStream:
        buffer = BytesIO()
        self._client.download_fileobj(
            Bucket=self._coords.bucket, Key=identifier["key"], Fileobj=buffer
        )
        buffer.seek(0)
        return DocumentStream(name=identifier["key"], stream=buffer)

    def _fetch_documents(self):
        for key in self._list_document_ids():
            yield self._fetch_document_by_id(key)



import logging
import shutil
import tempfile
from pathlib import Path
from typing import Any, Optional, Union

import msgpack
from rq import SimpleWorker, get_current_job

from docling.datamodel.base_models import DocumentStream

from docling_jobkit.convert.chunking import process_chunk_results
from docling_jobkit.convert.manager import (
    DoclingConverterManager,
    DoclingConverterManagerConfig,
)
from docling_jobkit.convert.results import process_export_results
from docling_jobkit.datamodel.http_inputs import FileSource, HttpSource
from docling_jobkit.datamodel.result import DoclingTaskResult
from docling_jobkit.datamodel.task import Task
from docling_jobkit.datamodel.task_meta import TaskStatus, TaskType
from docling_jobkit.orchestrators.rq.orchestrator import (
    RQOrchestrator,
    RQOrchestratorConfig,
    _TaskUpdate,
)

_log = logging.getLogger(__name__)


def make_msgpack_safe(obj):
    """
    Recursively convert any non-msgpack-serializable types to safe types,
    keeping bytes unchanged.
    """
    from datetime import datetime
    from decimal import Decimal

    # Types msgpack already supports
    if obj is None or isinstance(obj, (str, int, float, bool, bytes)):
        return obj

    # Handle sequences
    if isinstance(obj, (list, tuple, set)):
        return [make_msgpack_safe(v) for v in obj]

    # Handle mappings
    if isinstance(obj, dict):
        return {make_msgpack_safe(k): make_msgpack_safe(v) for k, v in obj.items()}

    # Known common conversions
    if isinstance(obj, (datetime, Decimal)):
        return str(obj)  # ISO for datetime, str for Decimal

    # Fallback: use string representation
    return str(obj)


class CustomRQWorker(SimpleWorker):
    def __init__(
        self,
        *args,
        orchestrator_config: RQOrchestratorConfig,
        cm_config: DoclingConverterManagerConfig,
        scratch_dir: Path,
        **kwargs,
    ):
        self.orchestrator_config = orchestrator_config
        self.conversion_manager = DoclingConverterManager(cm_config)
        self.scratch_dir = scratch_dir

        if "default_result_ttl" not in kwargs:
            kwargs["default_result_ttl"] = self.orchestrator_config.results_ttl

        # Call parent class constructor
        super().__init__(*args, **kwargs)

    def perform_job(self, job, queue):
        try:
            # Add to job's kwargs conversion manager
            if hasattr(job, "kwargs"):
                job.kwargs["conversion_manager"] = self.conversion_manager
                job.kwargs["orchestrator_config"] = self.orchestrator_config
                job.kwargs["scratch_dir"] = self.scratch_dir

            return super().perform_job(job, queue)
        except Exception as e:
            # Custom error handling for individual jobs
            self.log.error(f"Job {job.id} failed: {e}")
            raise


def docling_task(
    task_data: dict,
    conversion_manager: DoclingConverterManager,
    orchestrator_config: RQOrchestratorConfig,
    scratch_dir: Path,
):
    _log.debug("started task")
    task = Task.model_validate(task_data)
    task_id = task.task_id

    job = get_current_job()
    assert job is not None
    conn = job.connection

    # Notify task status
    conn.publish(
        orchestrator_config.sub_channel,
        _TaskUpdate(
            task_id=task_id,
            task_status=TaskStatus.STARTED,
        ).model_dump_json(),
    )

    workdir = scratch_dir / task_id

    try:
        _log.debug(f"task_id inside task is: {task_id}")
        convert_sources: list[Union[str, DocumentStream]] = []
        headers: Optional[dict[str, Any]] = None
        for source in task.sources:
            if isinstance(source, DocumentStream):
                convert_sources.append(source)
            elif isinstance(source, FileSource):
                convert_sources.append(source.to_document_stream())
            elif isinstance(source, HttpSource):
                convert_sources.append(str(source.url))
                if headers is None and source.headers:
                    headers = source.headers

        if not conversion_manager:
            raise RuntimeError("No converter")
        if not task.convert_options:
            raise RuntimeError("No conversion options")
        conv_results = conversion_manager.convert_documents(
            sources=convert_sources,
            options=task.convert_options,
            headers=headers,
        )

        processed_results: DoclingTaskResult
        if task.task_type == TaskType.CONVERT:
            processed_results = process_export_results(
                task=task,
                conv_results=conv_results,
                work_dir=workdir,
            )
        elif task.task_type == TaskType.CHUNK:
            processed_results = process_chunk_results(
                task=task,
                conv_results=conv_results,
                work_dir=workdir,
            )
        safe_data = make_msgpack_safe(processed_results.model_dump())
        packed = msgpack.packb(safe_data, use_bin_type=True)
        result_key = f"{orchestrator_config.results_prefix}:{task_id}"
        conn.setex(result_key, orchestrator_config.results_ttl, packed)

        # Notify task status
        conn.publish(
            orchestrator_config.sub_channel,
            _TaskUpdate(
                task_id=task_id,
                task_status=TaskStatus.SUCCESS,
                result_key=result_key,
            ).model_dump_json(),
        )

        _log.debug("ended task")
    except Exception as e:
        _log.error(f"Conversion task {task_id} failed: {e}")
        # Notify task status
        conn.publish(
            orchestrator_config.sub_channel,
            _TaskUpdate(
                task_id=task_id,
                task_status=TaskStatus.FAILURE,
            ).model_dump_json(),
        )
        raise e

    finally:
        if workdir.exists():
            shutil.rmtree(workdir)

    return result_key


def run_worker(
    rq_config: Optional[RQOrchestratorConfig] = None,
    cm_config: Optional[DoclingConverterManagerConfig] = None,
):
    # create a new connection in thread, Redis and ConversionManager are not pickle
    rq_config = rq_config or RQOrchestratorConfig()
    scratch_dir = rq_config.scratch_dir or Path(tempfile.mkdtemp(prefix="docling_"))
    redis_conn, rq_queue = RQOrchestrator.make_rq_queue(rq_config)
    cm_config = cm_config or DoclingConverterManagerConfig()
    worker = CustomRQWorker(
        [rq_queue],
        connection=redis_conn,
        orchestrator_config=rq_config,
        cm_config=cm_config,
        scratch_dir=scratch_dir,
    )
    worker.work()


if __name__ == "__main__":
    run_worker()



from abc import ABC, abstractmethod
from contextlib import AbstractContextManager
from itertools import islice
from typing import Callable, Generic, Iterator, Sequence, TypeVar

from docling.datamodel.base_models import DocumentStream

FileIdentifierT = TypeVar("FileIdentifierT")  # identifier type per connector


class DocumentChunk(Generic[FileIdentifierT]):
    def __init__(
        self,
        ids: Sequence[FileIdentifierT],
        fetcher: Callable[[FileIdentifierT], DocumentStream],
        chunk_index: int,
    ):
        self.ids = ids
        self._fetcher = fetcher
        self.index = chunk_index

    def iter_documents(self) -> Iterator[DocumentStream]:
        for doc_id in self.ids:
            yield self._fetcher(doc_id)


class BaseSourceProcessor(Generic[FileIdentifierT], AbstractContextManager, ABC):
    """
    Base class for source processors.
    Handles initialization state and context management.
    """

    def __init__(self):
        self._initialized = False  # Track whether the processor is ready

    def __enter__(self):
        self._initialize()
        self._initialized = True
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._finalize()
        self._initialized = False

    @abstractmethod
    def _initialize(self):
        """Prepare the processor (authenticate, open SDK clients, etc.)."""

    @abstractmethod
    def _finalize(self):
        """Clean up resources."""

    @abstractmethod
    def _fetch_documents(self) -> Iterator[DocumentStream]:
        """Yield documents from the source."""

    def _list_document_ids(self) -> Iterator[FileIdentifierT] | None:
        return None

    def _fetch_document_by_id(self, identifier: FileIdentifierT) -> DocumentStream:
        raise NotImplementedError

    def _count_documents(self) -> int | None:
        return None

    def iterate_documents(self) -> Iterator[DocumentStream]:
        if not self._initialized:
            raise RuntimeError(
                "Processor not initialized. Use 'with' to open it first."
            )
        yield from self._fetch_documents()

    def iterate_document_chunks(
        self, chunk_size: int
    ) -> Iterator[DocumentChunk[FileIdentifierT]]:
        ids_gen = self._list_document_ids()
        if ids_gen is None:
            raise RuntimeError("Connector does not support chunking.")

        chunk_index = 0

        while True:
            ids = list(islice(ids_gen, chunk_size))
            if not ids:
                break

            yield DocumentChunk(
                ids=ids,
                fetcher=self._fetch_document_by_id,
                chunk_index=chunk_index,
            )

            chunk_index += 1



from docling_jobkit.connectors.google_drive_target_processor import (
    GoogleDriveTargetProcessor,
)
from docling_jobkit.connectors.local_path_target_processor import (
    LocalPathTargetProcessor,
)
from docling_jobkit.connectors.s3_target_processor import S3TargetProcessor
from docling_jobkit.connectors.target_processor import BaseTargetProcessor
from docling_jobkit.datamodel.task_targets import (
    GoogleDriveTarget,
    LocalPathTarget,
    S3Target,
    TaskTarget,
)


def get_target_processor(target: TaskTarget) -> BaseTargetProcessor:
    if isinstance(target, S3Target):
        return S3TargetProcessor(target)
    if isinstance(target, GoogleDriveTarget):
        return GoogleDriveTargetProcessor(target)
    if isinstance(target, LocalPathTarget):
        return LocalPathTargetProcessor(target)

    raise RuntimeError(f"No target processor for this target. {type(target)=}")



from typing import Annotated, List, Optional

from pydantic import BaseModel, Field, HttpUrl, SecretStr, StrictStr


class GoogleDriveCredentials(BaseModel):
    client_id: Annotated[
        StrictStr,
        Field(
            description="OAuth 2.0 Client ID issued by Google. Required.",
        ),
    ]

    project_id: Annotated[
        StrictStr,
        Field(
            description="Google Cloud project ID associated with the OAuth client. Required.",
            examples=["docling-test-473014"],
        ),
    ]

    auth_uri: Annotated[
        HttpUrl,
        Field(
            description="Authorization endpoint URI. Required.",
            examples=["https://accounts.google.com/o/oauth2/auth"],
        ),
    ]

    token_uri: Annotated[
        HttpUrl,
        Field(
            description="Token endpoint URI. Required.",
            examples=["https://oauth2.googleapis.com/token"],
        ),
    ]

    auth_provider_x509_cert_url: Annotated[
        HttpUrl,
        Field(
            description="Certs URL for Google's OAuth provider. Required.",
            examples=["https://www.googleapis.com/oauth2/v1/certs"],
        ),
    ]

    client_secret: Annotated[
        SecretStr,
        Field(
            description="OAuth 2.0 client secret. Required.",
        ),
    ]

    redirect_uris: Annotated[
        List[HttpUrl],
        Field(
            description="OAuth 2.0 redirect URIs. Required.",
            examples=[["http://localhost"]],
        ),
    ]


class GoogleDriveCoordinates(BaseModel):
    path_id: Annotated[
        StrictStr,
        Field(
            description=(
                "Identifier for a file or folder in Google Drive. It can be obtained from the URL as follows:"
                "Folder: https://drive.google.com/drive/u/0/folders/1yucgL9WGgWZdM1TOuKkeghlPizuzMYb5 -> folder id is 1yucgL9WGgWZdM1TOuKkeghlPizuzMYb5"
                "File: https://docs.google.com/document/d/1bfaMQ18_i56204VaQDVeAFpqEijJTgvurupdEDiaUQw/edit -> document id is 1bfaMQ18_i56204VaQDVeAFpqEijJTgvurupdEDiaUQw."
                "Required."
            ),
            examples=[
                "11hgbUnDr-fyX4Hsi3T2q3xvYimvkOrfN",
            ],
        ),
    ]

    token_path: Annotated[
        Optional[StrictStr],
        Field(
            default=None,
            description=(
                "Path to save the OAuth 2.0 access token, which is generated on the fly. One of 'token_path' or 'refresh_token' is required."
            ),
            examples=[
                "./dev/google_drive_token.json",
            ],
        ),
    ]

    refresh_token: Annotated[
        Optional[StrictStr],
        Field(
            default=None,
            description=(
                "Refresh token for the OAuth 2.0 access, if already pre-generated. One of 'token_path' or 'refresh_token' is required."
            ),
        ),
    ]

    credentials_path: Annotated[
        Optional[StrictStr],
        Field(
            default=None,
            description=(
                "Path to the OAuth 2.0 Client ID credentials (available in Google Cloud console). One of 'credentials_path' or 'credentials' is required."
            ),
            examples=[
                "./dev/google_drive_credentials.json",
            ],
        ),
    ]

    credentials: Annotated[
        Optional[GoogleDriveCredentials],
        Field(
            default=None,
            description="OAuth 2.0 Client ID' credentials (available in Google Cloud console). One of 'credentials_path' or 'credentials' is required.",
        ),
    ]



import logging
from pathlib import Path
from urllib.parse import urlunsplit

from boto3.session import Session
from botocore.config import Config
from mypy_boto3_s3 import S3Client
from mypy_boto3_s3.paginator import ListObjectsV2Paginator
from mypy_boto3_s3.service_resource import S3ServiceResource

from docling_jobkit.datamodel.s3_coords import S3Coordinates

logging.basicConfig(level=logging.INFO)

# Set the maximum file size of parquet to 500MB
MAX_PARQUET_FILE_SIZE = 500 * 1024 * 1024

classifier_labels = [
    "bar_chart",
    "bar_code",
    "chemistry_markush_structure",
    "chemistry_molecular_structure",
    "flow_chart",
    "icon",
    "line_chart",
    "logo",
    "map",
    "other",
    "pie_chart",
    "qr_code",
    "remote_sensing",
    "screenshot",
    "signature",
    "stamp",
]


def get_s3_connection(coords: S3Coordinates):
    session = Session()

    config = Config(
        connect_timeout=30, retries={"max_attempts": 1}, signature_version="s3v4"
    )
    scheme = "https" if coords.verify_ssl else "http"
    path = "/"
    endpoint = urlunsplit((scheme, coords.endpoint, path, "", ""))

    client: S3Client = session.client(
        "s3",
        endpoint_url=endpoint,
        verify=coords.verify_ssl,
        aws_access_key_id=coords.access_key.get_secret_value(),
        aws_secret_access_key=coords.secret_key.get_secret_value(),
        config=config,
    )

    resource: S3ServiceResource = session.resource(
        "s3",
        endpoint_url=endpoint,
        verify=coords.verify_ssl,
        aws_access_key_id=coords.access_key.get_secret_value(),
        aws_secret_access_key=coords.secret_key.get_secret_value(),
        config=config,
    )

    return client, resource


def count_s3_objects(paginator: ListObjectsV2Paginator, bucket_name: str, prefix: str):
    response_iterator = paginator.paginate(Bucket=bucket_name, Prefix=prefix)
    count_obj = 0
    for page in response_iterator:
        if page.get("Contents"):
            count_obj += sum(1 for _ in page["Contents"])

    return count_obj


def get_keys_s3_objects_as_set(
    s3_resource: S3ServiceResource, bucket_name: str, prefix: str
) -> set[str]:
    bucket = s3_resource.Bucket(bucket_name)
    folder_objects = list(bucket.objects.filter(Prefix=prefix))
    files_on_s3 = set()
    for file in folder_objects:
        files_on_s3.add(file.key)
    return files_on_s3


def strip_prefix_postfix(source_set: set[str], prefix: str = "", extension: str = ""):
    output = set()
    for key in source_set:
        output.add(key.replace(extension, "").replace(prefix, ""))
    return output


def generate_batch_keys(
    source_keys: list[str],
    batch_size: int = 10,
):
    batched_keys = []
    counter = 0
    sub_array = []
    array_lenght = len(source_keys)
    for idx, key in enumerate(source_keys):
        sub_array.append(key)
        counter += 1
        if counter == batch_size or (idx + 1) == array_lenght:
            batched_keys.append(sub_array)
            sub_array = []
            counter = 0

    return batched_keys


# TODO: raised default expiration_time raised due to presign being generated
# in compute batches with new convert manager. This probably is not be enough
def generate_presign_url(
    client: S3Client,
    object_key: str,
    bucket: str,
    expiration_time: int = 21600,
) -> str | None:
    try:
        return client.generate_presigned_url(
            ClientMethod="get_object",
            Params={"Bucket": bucket, "Key": object_key},
            ExpiresIn=expiration_time,
        )
    except Exception as e:
        logging.error("Generation of presigned url failed", exc_info=e)
        return None


def get_source_files(
    s3_source_client: S3Client,
    s3_source_resource: S3ServiceResource,
    s3_coords: S3Coordinates,
):
    source_paginator = s3_source_client.get_paginator("list_objects_v2")

    key_prefix = (
        s3_coords.key_prefix
        if s3_coords.key_prefix.endswith("/")
        else s3_coords.key_prefix + "/"
    )
    if key_prefix == "/":
        key_prefix = ""
    # Check that source is not empty
    source_count = count_s3_objects(source_paginator, s3_coords.bucket, key_prefix)
    if source_count == 0:
        logging.error("No documents to process in the source s3 coordinates.")
    return get_keys_s3_objects_as_set(s3_source_resource, s3_coords.bucket, key_prefix)


def check_target_has_source_converted(
    coords: S3Coordinates,
    source_objects_list: list[str],
    s3_source_prefix: str,
):
    s3_target_client, s3_target_resource = get_s3_connection(coords)
    target_paginator = s3_target_client.get_paginator("list_objects_v2")

    converted_prefix = (
        coords.key_prefix + "json/"
        if coords.key_prefix.endswith("/")
        else coords.key_prefix + "/json/"
    )

    target_count = count_s3_objects(target_paginator, coords.bucket, converted_prefix)
    logging.debug("Target contains json objects: {}".format(target_count))
    if target_count != 0:
        logging.debug("Target contains objects, checking content...")

        # Collect target keys for iterative conversion
        existing_target_objects = get_keys_s3_objects_as_set(
            s3_target_resource, coords.bucket, converted_prefix
        )

        # At this point we should be targeting keys in the json "folder"
        target_short_key_list = []
        for item in existing_target_objects:
            clean_name = str(Path(item).stem)
            target_short_key_list.append(clean_name)

        filtered_source_keys = []
        logging.debug("List of source keys:")
        for key in source_objects_list:
            logging.debug("Object key: {}".format(key))
            # This covers the case when source docs have "folder" hierarchy in the key
            # we don't preserve key part between prefix and "file", this part of key is not added as prefix for target
            clean_key = str(Path(key).stem)
            if clean_key not in target_short_key_list:
                filtered_source_keys.append(key)

        logging.debug("Total keys: {}".format(len(source_objects_list)))
        logging.debug("Filtered keys to process: {}".format(len(filtered_source_keys)))
    else:
        filtered_source_keys = source_objects_list

    return filtered_source_keys



import warnings
from typing import Annotated, Literal, Optional

from pydantic import BaseModel, Field

from docling.datamodel.base_models import ConversionStatus, ErrorItem
from docling.utils.profiling import ProfilingItem
from docling_core.types.doc.document import DoclingDocument


class ExportDocumentResponse(BaseModel):
    filename: str
    md_content: Optional[str] = None
    json_content: Optional[DoclingDocument] = None
    html_content: Optional[str] = None
    text_content: Optional[str] = None
    doctags_content: Optional[str] = None


class ExportResult(BaseModel):
    """Container of all exported content."""

    kind: Literal["ExportResult"] = "ExportResult"
    content: ExportDocumentResponse
    status: ConversionStatus
    errors: list[ErrorItem] = []
    timings: dict[str, ProfilingItem] = {}


class ZipArchiveResult(BaseModel):
    """Container for a zip archive of the conversion."""

    kind: Literal["ZipArchiveResult"] = "ZipArchiveResult"
    content: bytes


class RemoteTargetResult(BaseModel):
    """No content, the result has been pushed to a remote target."""

    kind: Literal["RemoteTargetResult"] = "RemoteTargetResult"


class ChunkedDocumentResultItem(BaseModel):
    """A single chunk of a document with its metadata and content."""

    filename: str
    chunk_index: int
    text: Annotated[
        str,
        Field(
            description="The chunk text with structural context (headers, formatting)"
        ),
    ]
    raw_text: Annotated[
        str | None,
        Field(
            description="Raw chunk text without additional formatting or context",
        ),
    ] = None
    num_tokens: Annotated[
        int | None,
        Field(
            description="Number of tokens in the text, if the chunker is aware of tokens"
        ),
    ] = None
    headings: Annotated[
        list[str] | None, Field(description="List of headings for this chunk")
    ] = None
    captions: Annotated[
        list[str] | None,
        Field(
            description="List of captions for this chunk (e.g. for pictures and tables)",
        ),
    ] = None
    doc_items: Annotated[list[str], Field(description="List of doc items references")]
    page_numbers: Annotated[
        list[int] | None,
        Field(description="Page numbers where this chunk content appears"),
    ] = None
    metadata: Annotated[
        dict | None, Field(description="Additional metadata associated with this chunk")
    ] = None


class ChunkedDocumentResult(BaseModel):
    kind: Literal["ChunkedDocumentResponse"] = "ChunkedDocumentResponse"
    chunks: list[ChunkedDocumentResultItem]
    documents: list[ExportResult]
    chunking_info: Optional[dict] = None


ResultType = Annotated[
    ExportResult | ZipArchiveResult | RemoteTargetResult | ChunkedDocumentResult,
    Field(discriminator="kind"),
]


class DoclingTaskResult(BaseModel):
    result: ResultType
    processing_time: float
    num_converted: int
    num_succeeded: int
    num_failed: int


class ConvertDocumentResult(DoclingTaskResult):
    def __init__(self, *args, **kwargs):
        warnings.warn(
            "ConvertDocumentResult is deprecated and will be removed in a future version. "
            "Use DoclingTaskResult instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(*args, **kwargs)



from docling_jobkit.connectors.google_drive_source_processor import (
    GoogleDriveSourceProcessor,
)
from docling_jobkit.connectors.http_source_processor import HttpSourceProcessor
from docling_jobkit.connectors.local_path_source_processor import (
    LocalPathSourceProcessor,
)
from docling_jobkit.connectors.s3_source_processor import S3SourceProcessor
from docling_jobkit.connectors.source_processor import BaseSourceProcessor
from docling_jobkit.datamodel.task_sources import (
    TaskFileSource,
    TaskGoogleDriveSource,
    TaskHttpSource,
    TaskLocalPathSource,
    TaskS3Source,
    TaskSource,
)


def get_source_processor(source: TaskSource) -> BaseSourceProcessor:
    if isinstance(source, (TaskFileSource, TaskHttpSource)):
        return HttpSourceProcessor(source)
    elif isinstance(source, TaskS3Source):
        return S3SourceProcessor(source)
    elif isinstance(source, TaskGoogleDriveSource):
        return GoogleDriveSourceProcessor(source)
    elif isinstance(source, TaskLocalPathSource):
        return LocalPathSourceProcessor(source)

    raise RuntimeError(f"No source processor for this source. {type(source)=}")



import json
import logging
from io import BytesIO
from pathlib import Path, PurePosixPath
from typing import Iterable, List, TypedDict

from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import Resource, build
from googleapiclient.http import MediaFileUpload, MediaIoBaseDownload, MediaIoBaseUpload

from docling_jobkit.datamodel.google_drive_coords import GoogleDriveCoordinates


class GoogleDriveFileIdentifier(TypedDict):
    id: str
    name: str
    mimeType: str
    path: str


def _get_client_info(coords):
    if coords.credentials:
        return (
            coords.credentials.client_id,
            coords.credentials.client_secret.get_secret_value(),
            str(coords.credentials.token_uri),
        )
    if coords.credentials_path:
        data = json.loads(Path(coords.credentials_path).read_text())
        conf = data.get("installed")
        return conf["client_id"], conf["client_secret"], conf["token_uri"]
    raise ValueError("Missing client credentials")


def get_service(coords: GoogleDriveCoordinates) -> Resource:
    """
    Return an authorized Google Drive service (googleapiclient.discovery.Resource). Build the service using, in priority:
      1) An already-stored token file,
      2) An already-generated refresh token,
      3) The local OAuth flow.
    """
    SCOPES = ["https://www.googleapis.com/auth/drive"]
    creds = None

    # 1) Stored token
    if coords.token_path and Path(coords.token_path).exists():
        creds = Credentials.from_authorized_user_file(coords.token_path, SCOPES)
        if creds.expired and creds.refresh_token:
            creds.refresh(Request())

    # 2) Refresh token
    if (not creds or not creds.valid) and coords.refresh_token:
        try:
            client_id, client_secret, token_uri = _get_client_info(coords)
            creds = Credentials(
                token=None,
                refresh_token=coords.refresh_token,
                token_uri=token_uri,
                client_id=client_id,
                client_secret=client_secret,
                scopes=SCOPES,
            )
            creds.refresh(Request())
        except Exception as e:
            logging.warning("Refresh failed: %s", e)
            creds = None

    # 3) OAuth flow
    if not creds or not creds.valid:
        if coords.credentials:
            cfg = coords.credentials.model_dump(mode="json", exclude_none=True)
            cfg["client_secret"] = coords.credentials.client_secret.get_secret_value()
            flow = InstalledAppFlow.from_client_config({"installed": cfg}, SCOPES)
        else:
            flow = InstalledAppFlow.from_client_secrets_file(
                coords.credentials_path, SCOPES
            )

        creds = flow.run_local_server(port=0)

    # Save token
    if coords.token_path:
        Path(coords.token_path).write_text(creds.to_json())

    return build("drive", "v3", credentials=creds, cache_discovery=False)


def _yield_children(service: Resource, folder_id: str):
    """Yield direct children of a folder."""

    query = f"'{folder_id}' in parents and trashed = false"
    fields = (
        "nextPageToken, "
        "files(id, name, mimeType, parents, shortcutDetails(targetId, targetMimeType))"
    )
    page_token = None
    while True:
        resp = (
            service.files()
            .list(
                q=query,
                spaces="drive",
                fields=fields,
                pageToken=page_token,
                pageSize=1000,
                includeItemsFromAllDrives=True,
                supportsAllDrives=True,
            )
            .execute()
        )
        for f in resp.get("files", []):
            yield f
        page_token = resp.get("nextPageToken")
        if not page_token:
            break


def _yield_files_infos(
    service: Resource,
    coords: GoogleDriveCoordinates,
) -> Iterable[GoogleDriveFileIdentifier]:
    """
    Depth-first traversal of Google Drive.
    Yields dicts: {id, name, mimeType, path}
    """

    root_meta = (
        service.files()
        .get(
            fileId=coords.path_id,
            fields="id, name, mimeType",
            supportsAllDrives=True,
        )
        .execute()
    )

    info: GoogleDriveFileIdentifier
    if not (root_meta.get("mimeType") == "application/vnd.google-apps.folder"):
        info = {
            "id": root_meta["id"],
            "name": root_meta["name"],
            "mimeType": root_meta["mimeType"],
            "path": root_meta["name"],
        }
        yield info
        return

    stack = [(coords.path_id, root_meta["name"])]
    while stack:
        cur_id, cur_path = stack.pop()
        for item in _yield_children(service, cur_id):
            path = f"{cur_path}/{item['name']}"
            if item["mimeType"] == "application/vnd.google-apps.folder":
                stack.append((item["id"], path))
            else:
                info = {
                    "id": item["id"],
                    "name": item["name"],
                    "mimeType": item["mimeType"],
                    "path": path,
                }
                yield info


def get_source_files_infos(
    service: Resource,
    coords: GoogleDriveCoordinates,
) -> List[GoogleDriveFileIdentifier]:
    return list(_yield_files_infos(service, coords))


def download_file(
    service: Resource,
    file_info: GoogleDriveFileIdentifier,
    file_stream: BytesIO,
) -> None:
    """
    Download a file from Google Drive.
    The file can be any file stored in Google Drive as well as a document created with Google Slides, Google Sheets or Google Docs.
    """

    EXPORT_MAP = {
        "application/vnd.google-apps.document": (
            "application/vnd.openxmlformats-officedocument.wordprocessingml.document"
        ),
        "application/vnd.google-apps.spreadsheet": (
            "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"
        ),
        "application/vnd.google-apps.presentation": (
            "application/vnd.openxmlformats-officedocument.presentationml.presentation"
        ),
    }
    if file_info["mimeType"].startswith("application/vnd.google-apps."):
        export_mime = EXPORT_MAP.get(file_info["mimeType"])
        request = service.files().export_media(
            fileId=file_info["id"], mimeType=export_mime
        )
    else:
        request = service.files().get_media(
            fileId=file_info["id"], supportsAllDrives=True
        )

    downloader = MediaIoBaseDownload(file_stream, request)
    done = False
    while not done:
        status, done = downloader.next_chunk()
        if status:
            logging.info("Downloading: %d%%", int(status.progress() * 100))


def _create_subfolders(service, base_folder_id: str, target_path_parent: str) -> str:
    """
    Create subfolders in "target_path_parent". For instance, if "target_path_parent" is "pages/abc.png", the folder "pages" is created in Google Drive.
    Returns the id of the deepest folder of "relative_path".
    """
    deepest_folder_id = base_folder_id
    for part in PurePosixPath(target_path_parent).parts:
        escaped_name = part.replace("\\", "\\\\").replace("'", "\\'")
        resp = (
            service.files()
            .list(
                q=(
                    f"'{deepest_folder_id}' in parents and "
                    f"name = '{escaped_name}' and "
                    f"mimeType = 'application/vnd.google-apps.folder' and trashed = false"
                ),
                fields="files(id)",
                pageSize=1,
                supportsAllDrives=True,
                includeItemsFromAllDrives=True,
                spaces="drive",
            )
            .execute()
        )
        files = resp.get("files", [])
        if files:
            deepest_folder_id = files[0]["id"]
        else:
            # Create folder
            created = (
                service.files()
                .create(
                    body={
                        "name": part,
                        "mimeType": "application/vnd.google-apps.folder",
                        "parents": [deepest_folder_id],
                    },
                    fields="id",
                    supportsAllDrives=True,
                )
                .execute()
            )
            deepest_folder_id = created["id"]
    return deepest_folder_id


def upload_file(
    service,
    target_filename: str,
    content_type: str,
    coords,
    filename=None,
    file_stream=None,
) -> None:
    """Upload a file to Google Drive; auto-create subfolders from target_filename."""

    # Check that "coords.path_id" is a folder
    meta = (
        service.files()
        .get(
            fileId=coords.path_id,
            fields="id, mimeType",
            supportsAllDrives=True,
        )
        .execute()
    )
    if meta.get("mimeType") != "application/vnd.google-apps.folder":
        raise ValueError(f"path_id must be a folder, got {meta.get('mimeType')}")

    # If "target_filename" contains subfolders, create them and find the deepest folder id.
    target_path = PurePosixPath(target_filename)
    if str(target_path.parent) and str(target_path.parent) != ".":
        parent_id = _create_subfolders(service, coords.path_id, str(target_path.parent))
    else:
        parent_id = coords.path_id

    # Upload file
    chunk_size = 8 * 1024 * 1024
    if file_stream is not None:
        media = MediaIoBaseUpload(
            file_stream, mimetype=content_type, chunksize=chunk_size, resumable=True
        )
    else:
        media = MediaFileUpload(
            filename, mimetype=content_type, chunksize=chunk_size, resumable=True
        )
    request = service.files().create(
        body={"name": target_path.name, "parents": [parent_id]},
        media_body=media,
        fields="id, name, webViewLink",
        supportsAllDrives=True,
    )
    response = None
    while response is None:
        status, response = request.next_chunk()
        if status:
            prog = getattr(status, "resumable_progress", None)
            if prog is not None:
                logging.info("uploading %s: %d bytes sent", filename, prog)
            else:
                logging.info("uploading %s...", filename)

    logging.info(
        "uploaded %s (%s) → %s",
        response.get("name"),
        response.get("id"),
        response.get("webViewLink"),
    )



import enum
from typing import Annotated, Literal, Optional

from pydantic import BaseModel, Field


class ChunkingExportOptions(BaseModel):
    """Export options for the response of the chunking task."""

    include_converted_doc: bool = False


class ChunkerType(str, enum.Enum):
    """Choice of the chunkers available in Docling."""

    HIERARCHICAL = "hierarchical"
    HYBRID = "hybrid"


class BaseChunkerOptions(BaseModel):
    """Configuration options for document chunking using Docling chunkers."""

    chunker: ChunkerType

    use_markdown_tables: Annotated[
        bool,
        Field(
            description="Use markdown table format instead of triplets for table serialization.",
        ),
    ] = False

    include_raw_text: Annotated[
        bool,
        Field(
            description="Include both raw_text and text (contextualized) in response. If False, only text is included.",
        ),
    ] = False


class HierarchicalChunkerOptions(BaseChunkerOptions):
    """Configuration options for the HierarchicalChunker."""

    chunker: Literal[ChunkerType.HIERARCHICAL] = ChunkerType.HIERARCHICAL


class HybridChunkerOptions(BaseChunkerOptions):
    """Configuration options for the HybridChunker."""

    chunker: Literal[ChunkerType.HYBRID] = ChunkerType.HYBRID

    max_tokens: Annotated[
        Optional[int],
        Field(
            description="Maximum number of tokens per chunk. When left to none, the value is automatically extracted from the tokenizer.",
        ),
    ] = None

    tokenizer: Annotated[
        str,
        Field(
            description="HuggingFace model name for custom tokenization. If not specified, uses 'sentence-transformers/all-MiniLM-L6-v2' as default.",
            examples=[
                "Qwen/Qwen3-Embedding-0.6B",
                "sentence-transformers/all-MiniLM-L6-v2",
            ],
        ),
    ] = "sentence-transformers/all-MiniLM-L6-v2"

    merge_peers: Annotated[
        bool,
        Field(
            description="Merge undersized successive chunks with same headings.",
        ),
    ] = True


# Union of all chunking options in the factory
ChunkingOptionType = HybridChunkerOptions | HierarchicalChunkerOptions



# ruff: noqa: E402

from typing import List, NamedTuple

from kfp import dsl, kubernetes
from kfp.dsl import Dataset, Input, Output


@dsl.component(
    packages_to_install=[
        "docling==2.50.0",
        "boto3~=1.35.36",
        "git+https://github.com/docling-project/docling-jobkit@main",
    ],
    base_image="ghcr.io/docling-project/docling-serve-cpu:v1.5.1",
)
def convert_payload(
    options: dict,
    source: dict,
    target: dict,
    batch_index: int,
    # source_keys: List[str],
    dataset: Input[Dataset],
) -> list:
    import json
    import logging
    from pathlib import Path

    from docling.datamodel.base_models import ConversionStatus, InputFormat
    from docling.datamodel.pipeline_options import (
        VlmPipelineOptions,
    )
    from docling.datamodel.pipeline_options_vlm_model import (
        ApiVlmOptions,
        ResponseFormat,
    )
    from docling.datamodel.settings import settings
    from docling.document_converter import DocumentConverter, PdfFormatOption
    from docling.pipeline.vlm_pipeline import VlmPipeline
    from docling_core.types.doc.base import ImageRefMode

    from docling_jobkit.connectors.s3_helper import (
        generate_presign_url,
        get_s3_connection,
    )
    from docling_jobkit.datamodel.s3_coords import S3Coordinates

    def openai_compatible_vlm_options(
        model: str,
        prompt: str,
        format: ResponseFormat,
        hostname_and_port,
        temperature: float = 0.7,
        max_tokens: int = 4096,
        api_key: str = "",
        skip_special_tokens=False,
    ):
        headers = {}
        if api_key:
            headers["Authorization"] = f"Bearer {api_key}"

        options = ApiVlmOptions(
            url=f"http://{hostname_and_port}/v1/chat/completions",  # LM studio defaults to port 1234, VLLM to 8000
            params={
                "model": model,
                "max_tokens": max_tokens,
                "skip_special_tokens": skip_special_tokens,  # needed for VLLM
            },
            headers=headers,
            prompt=prompt,
            timeout=90,
            scale=2.0,
            temperature=temperature,
            response_format=format,
        )
        return options

    logging.basicConfig(level=logging.INFO)
    settings.debug.profile_pipeline_timings = True

    # validate coords
    s3_coords_source = S3Coordinates.model_validate(source)
    target_s3_coords = S3Coordinates.model_validate(target)
    s3_source_client, s3_source_resource = get_s3_connection(s3_coords_source)
    s3_target_client, s3_target_resource = get_s3_connection(target_s3_coords)

    with open(dataset.path) as f:
        batches = json.load(f)
    source_keys = batches[batch_index]

    presign_filtered_source_keys = [
        generate_presign_url(s3_source_client, key, s3_coords_source.bucket)
        for key in source_keys
    ]

    pipeline_options = VlmPipelineOptions(
        enable_remote_services=True  # required when calling remote VLM endpoints
    )

    pipeline_options.vlm_options = openai_compatible_vlm_options(
        model="ibm-granite/granite-docling-258M",  # For VLLM use "ibm-granite/granite-docling-258M"
        hostname_and_port="vllm-gpu.deep-search.svc.cluster.local:8000",  # LM studio defaults to port 1234, VLLM to 8000
        prompt="Convert this page to docling.",
        format=ResponseFormat.DOCTAGS,
        api_key="",
    )

    doc_converter = DocumentConverter(
        format_options={
            InputFormat.PDF: PdfFormatOption(
                pipeline_options=pipeline_options,
                pipeline_cls=VlmPipeline,
            )
        }
    )

    for item in presign_filtered_source_keys:
        conv_res = doc_converter.convert(str(item))
        if conv_res.status == ConversionStatus.SUCCESS:
            filename = conv_res.input.document_hash

            # store doc
            temp_folder = Path("/modelcache/docling-temp")
            temp_folder.mkdir(parents=True, exist_ok=True)
            temp_json_file = temp_folder / f"{filename}.temp.json"
            conv_res.document.save_as_json(
                filename=temp_json_file,
                image_mode=ImageRefMode.EMBEDDED,
            )
            kwargs = {}
            kwargs["ContentType"] = "application/json"
            try:
                s3_target_client.upload_file(
                    Filename=temp_json_file,
                    Bucket=target_s3_coords.bucket,
                    Key=f"{target_s3_coords.key_prefix}/{filename}.json",
                    ExtraArgs={**kwargs},
                )
                # s3_target_client.put_object(
                #     Body=json.dumps(conv_res.document.export_to_dict()),
                #     Bucket=target_s3_coords.bucket,
                #     Key=f"{target_s3_coords.key_prefix}/{filename}.json",
                #     ContentType='"application/json"'
                # )
            except Exception as e:
                logging.error("Uploading document to s3 failed: {}".format(e))

            # clean-up
            temp_json_file.unlink()

            # store timings
            timings = {}
            for key in conv_res.timings.keys():
                timings[key] = {
                    "scope": conv_res.timings[key].scope.name,
                    "count": conv_res.timings[key].count,
                    "times": conv_res.timings[key].times,
                }
            try:
                s3_target_client.put_object(
                    Body=json.dumps(timings),
                    Bucket=target_s3_coords.bucket,
                    Key=f"{target_s3_coords.key_prefix}/{filename}.timings.json",
                    ContentType="application/json",
                )
            except Exception as e:
                logging.error("Uploading timings to s3 failed: {}".format(e))

    return []


@dsl.component(
    packages_to_install=[
        "pydantic",
        "boto3~=1.35.36",
        "git+https://github.com/docling-project/docling-jobkit@main",
    ],
    base_image="ghcr.io/docling-project/docling-serve-cpu:v1.5.1",
)
def compute_batches(
    source: dict,
    target: dict,
    dataset: Output[Dataset],
    batch_size: int = 10,
) -> NamedTuple("outputs", [("batch_indices", List[int])]):  # type: ignore[valid-type]
    import json
    import logging
    from typing import NamedTuple

    from docling_jobkit.connectors.s3_helper import (
        check_target_has_source_converted,
        generate_batch_keys,
        get_s3_connection,
        get_source_files,
    )
    from docling_jobkit.datamodel.s3_coords import S3Coordinates

    logging.basicConfig(level=logging.INFO)

    # validate inputs
    s3_coords_source = S3Coordinates.model_validate(source)
    s3_target_coords = S3Coordinates.model_validate(target)

    s3_source_client, s3_source_resource = get_s3_connection(s3_coords_source)
    source_objects_list = get_source_files(
        s3_source_client, s3_source_resource, s3_coords_source
    )
    filtered_source_keys = check_target_has_source_converted(
        s3_target_coords, source_objects_list, s3_coords_source.key_prefix
    )
    batch_keys = generate_batch_keys(
        filtered_source_keys,
        batch_size=batch_size,
    )

    # store batches on s3 for debugging
    s3_target_client, s3_target_resource = get_s3_connection(s3_target_coords)
    try:
        s3_target_client.put_object(
            Body=json.dumps(batch_keys),
            Bucket=s3_target_coords.bucket,
            Key=f"{s3_target_coords.key_prefix}/buckets.json",
            ContentType="application/json",
        )
    except Exception as e:
        logging.error("Uploading batch.json to s3 failed: {}".format(e))

    with open(dataset.path, "w") as out_batches:
        json.dump(batch_keys, out_batches)

    batch_indices = list(range(len(batch_keys)))
    outputs = NamedTuple("outputs", [("batch_indices", List[int])])
    print(f"Batches to convert: {len(batch_keys)}")
    return outputs(batch_indices)


@dsl.pipeline
def docling_s3in_s3out_with_infer(
    convertion_options: dict = {
        "from_formats": [
            "docx",
            "pptx",
            "html",
            "image",
            "pdf",
            "asciidoc",
            "md",
            "xlsx",
            "xml_uspto",
            "xml_jats",
            "json_docling",
        ],
        "to_formats": ["md", "json", "html", "text", "doctags"],
        "image_export_mode": "placeholder",
        "do_ocr": True,
        "force_ocr": False,
        "ocr_engine": "easyocr",
        "ocr_lang": [],
        "pdf_backend": "dlparse_v2",
        "table_mode": "accurate",
        "abort_on_error": False,
        "return_as_file": False,
        "do_table_structure": True,
        "do_code_enrichment": False,
        "do_formula_enrichment": False,
        "do_picture_classification": False,
        "do_picture_description": False,
        "generate_picture_images": False,
        "include_images": True,
        "images_scale": 2,
    },
    source: dict = {
        "endpoint": "s3.eu-de.cloud-object-storage.appdomain.cloud",
        "access_key": "123454321",
        "secret_key": "secretsecret",
        "bucket": "source-bucket",
        "key_prefix": "my-docs",
        "verify_ssl": True,
    },
    target: dict = {
        "endpoint": "s3.eu-de.cloud-object-storage.appdomain.cloud",
        "access_key": "123454321",
        "secret_key": "secretsecret",
        "bucket": "target-bucket",
        "key_prefix": "my-docs",
        "verify_ssl": True,
    },
    batch_size: int = 20,
):
    import logging

    logging.basicConfig(level=logging.INFO)

    batches = compute_batches(source=source, target=target, batch_size=batch_size)
    # disable caching on batches as cached pre-signed urls might have already expired
    batches.set_caching_options(False)

    results = []
    with dsl.ParallelFor(batches.outputs["batch_indices"], parallelism=20) as subbatch:
        converter = convert_payload(
            options=convertion_options,
            source=source,
            target=target,
            dataset=batches.outputs["dataset"],
            batch_index=subbatch,
        )
        converter.set_memory_request("1G")
        converter.set_memory_limit("16G")
        converter.set_cpu_request("200m")
        converter.set_cpu_limit("1")

        # All models should be preloaded to pvc, multi access pvc can be used by all pods in parallel
        kubernetes.mount_pvc(
            converter,
            pvc_name="docling-model-cache-pvc-multi",
            mount_path="/modelcache",
        )

        # For enabling document conversion using GPU
        # currently unable to properly pass input parameters into pipeline, therefore node selector and tolerations are hardcoded

        # converter.set_accelerator_type("nvidia.com/gpu")
        # converter.set_accelerator_limit("1")

        # kubernetes.add_node_selector(
        #     task=converter,
        #     label_key="nvidia.com/gpu.product",
        #     label_value="NVIDIA-A10",
        # )

        kubernetes.add_toleration(
            task=converter,
            key="key1",
            operator="Equal",
            value="mcad",
            effect="NoSchedule",
        )

        results.append(converter.output)


### Compile pipeline into a yaml
from kfp import compiler

compiler.Compiler().compile(
    docling_s3in_s3out_with_infer, "docling_s3in_s3out_with_infer.yaml"
)



from abc import ABC, abstractmethod
from contextlib import AbstractContextManager
from pathlib import Path
from typing import BinaryIO


class BaseTargetProcessor(AbstractContextManager, ABC):
    """
    Abstract base class for target processors that handle writing/uploading
    objects to a storage backend (e.g. S3, local FS, GCS).
    """

    def __init__(self):
        self._initialized = False

    def __enter__(self):
        self._initialize()
        self._initialized = True
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._finalize()
        self._initialized = False

    @abstractmethod
    def _initialize(self) -> None:
        """
        Subclasses must implement setup (e.g. create boto3 client).
        """
        ...

    @abstractmethod
    def _finalize(self) -> None:
        """
        Subclasses must implement teardown (e.g. close client).
        """
        ...

    @abstractmethod
    def upload_file(
        self,
        filename: str | Path,
        target_filename: str,
        content_type: str,
    ) -> None:
        """
        Upload a file from local filesystem.
        """
        ...

    @abstractmethod
    def upload_object(
        self,
        obj: str | bytes | BinaryIO,
        target_filename: str,
        content_type: str,
    ) -> None:
        """
        Upload an in-memory object (bytes or file-like) to the target.
        """
        ...



from pathlib import Path
from typing import Annotated, Optional

import typer
import yaml
from pydantic import BaseModel, Field, ValidationError
from rich.console import Console

from docling_jobkit.connectors.source_processor_factory import get_source_processor
from docling_jobkit.connectors.target_processor_factory import get_target_processor
from docling_jobkit.convert.manager import (
    DoclingConverterManager,
    DoclingConverterManagerConfig,
)
from docling_jobkit.convert.results_processor import ResultsProcessor
from docling_jobkit.datamodel.convert import ConvertDocumentsOptions
from docling_jobkit.datamodel.task_sources import (
    TaskFileSource,
    TaskGoogleDriveSource,
    TaskHttpSource,
    TaskLocalPathSource,
    TaskS3Source,
)
from docling_jobkit.datamodel.task_targets import (
    GoogleDriveTarget,
    LocalPathTarget,
    S3Target,
    ZipTarget,
)

console = Console()
err_console = Console(stderr=True)


app = typer.Typer(
    name="Docling Jobkit Local",
    no_args_is_help=True,
    add_completion=False,
    pretty_exceptions_enable=False,
)

JobTaskSource = Annotated[
    TaskFileSource
    | TaskHttpSource
    | TaskLocalPathSource
    | TaskS3Source
    | TaskGoogleDriveSource,
    Field(discriminator="kind"),
]

JobTaskTarget = Annotated[
    ZipTarget | LocalPathTarget | S3Target | GoogleDriveTarget,
    Field(discriminator="kind"),
]


class JobConfig(BaseModel):
    options: ConvertDocumentsOptions = ConvertDocumentsOptions()
    sources: list[JobTaskSource]
    target: JobTaskTarget


@app.command(no_args_is_help=True)
def convert(
    config_file: Annotated[
        Path,
        typer.Argument(
            help="Configuration file of the job", exists=True, readable=True
        ),
    ],
    artifacts_path: Annotated[
        Optional[Path],
        typer.Option(..., help="If provided, the location of the model artifacts."),
    ] = None,
    enable_remote_services: Annotated[
        bool,
        typer.Option(
            ..., help="Must be enabled when using models connecting to remote services."
        ),
    ] = False,
    allow_external_plugins: Annotated[
        bool,
        typer.Option(
            ..., help="Must be enabled for loading modules from third-party plugins."
        ),
    ] = False,
):
    # Open and validate config file
    try:
        with config_file.open("r") as f:
            raw_data = yaml.safe_load(f)
        config = JobConfig(**raw_data)
    except FileNotFoundError:
        typer.echo(f"❌ File not found: {config_file}")
    except ValidationError as e:
        typer.echo("❌ Validation failed:")
        typer.echo(e.json(indent=2))

    cm_config = DoclingConverterManagerConfig(
        artifacts_path=artifacts_path,
        enable_remote_services=enable_remote_services,
        allow_external_plugins=allow_external_plugins,
        options_cache_size=1,
    )
    manager = DoclingConverterManager(config=cm_config)

    results = []
    with get_target_processor(config.target) as target_processor:
        result_processor = ResultsProcessor(
            target_processor=target_processor,
            to_formats=[v.value for v in config.options.to_formats],
            generate_page_images=config.options.include_images,
            generate_picture_images=config.options.include_images,
        )
        for source in config.sources:
            with get_source_processor(source) as source_processor:
                for item in result_processor.process_documents(
                    manager.convert_documents(
                        sources=source_processor.iterate_documents(),
                        options=config.options,
                    )
                ):
                    results.append(item)


if __name__ == "__main__":
    app()



import json
import logging
import os
import shutil
import tempfile
from datetime import datetime
from io import BytesIO
from pathlib import Path
from typing import Any, Iterable

import pandas as pd
from pandas import DataFrame

from docling.datamodel.base_models import ConversionStatus
from docling.datamodel.document import ConversionResult
from docling.utils.utils import create_hash
from docling_core.types.doc.base import ImageRefMode
from docling_core.types.doc.document import (
    DocItem,
    DoclingDocument,
    PageItem,
    PictureClassificationData,
    PictureItem,
)
from docling_core.types.doc.labels import DocItemLabel

from docling_jobkit.connectors.s3_helper import (
    MAX_PARQUET_FILE_SIZE,
    classifier_labels,
)
from docling_jobkit.connectors.target_processor import BaseTargetProcessor


class ResultsProcessor:
    def __init__(
        self,
        target_processor: BaseTargetProcessor,
        to_formats: list[str] | None = None,
        generate_page_images: bool = False,
        generate_picture_images: bool = False,
        export_parquet_file: bool = False,
        scratch_dir: Path | None = None,
    ):
        self._target_processor = target_processor

        self.export_page_images = generate_page_images
        self.export_images = generate_picture_images

        self.to_formats = to_formats
        self.export_parquet_file = export_parquet_file

        self.scratch_dir = scratch_dir or Path(tempfile.mkdtemp(prefix="docling_"))
        self.scratch_dir.mkdir(exist_ok=True, parents=True)

    def __del__(self):
        if self.scratch_dir is not None:
            shutil.rmtree(self.scratch_dir, ignore_errors=True)

    def process_documents(self, results: Iterable[ConversionResult]):
        pd_d = DataFrame()  # DataFrame to append parquet info
        try:
            for i, conv_res in enumerate(results):
                with tempfile.TemporaryDirectory(dir=self.scratch_dir) as tmpdirname:
                    temp_dir = Path(tmpdirname)
                    if conv_res.status == ConversionStatus.SUCCESS:
                        doc_hash = conv_res.input.document_hash
                        name_without_ext = os.path.splitext(conv_res.input.file)[0]
                        logging.debug(f"Converted {doc_hash} now saving results")

                        if os.path.exists(conv_res.input.file):
                            self._target_processor.upload_file(
                                filename=Path(conv_res.input.file),
                                target_filename=f"pdf/{name_without_ext}.pdf",
                                content_type="application/pdf",
                            )

                        if self.export_page_images:
                            # Export pages images:
                            self.upload_page_images(
                                conv_res.document.pages,
                                conv_res.input.document_hash,
                            )

                        if self.export_images:
                            # Export pictures
                            self.upload_pictures(
                                conv_res.document,
                                conv_res.input.document_hash,
                            )

                        if self.to_formats is None or (
                            self.to_formats and "json" in self.to_formats
                        ):
                            # Export Docling document format to JSON:
                            target_key = f"json/{name_without_ext}.json"
                            temp_json_file = temp_dir / f"{name_without_ext}.json"

                            conv_res.document.save_as_json(
                                filename=temp_json_file,
                                image_mode=ImageRefMode.REFERENCED,
                            )
                            self._target_processor.upload_file(
                                filename=temp_json_file,
                                target_filename=target_key,
                                content_type="application/json",
                            )
                        if self.to_formats is None or (
                            self.to_formats and "doctags" in self.to_formats
                        ):
                            # Export Docling document format to doctags:
                            target_key = f"doctags/{name_without_ext}.doctags.txt"

                            data = conv_res.document.export_to_doctags()
                            self._target_processor.upload_object(
                                obj=data,
                                target_filename=target_key,
                                content_type="text/plain",
                            )
                        if self.to_formats is None or (
                            self.to_formats and "md" in self.to_formats
                        ):
                            # Export Docling document format to markdown:
                            target_key = f"md/{name_without_ext}.md"

                            data = conv_res.document.export_to_markdown()
                            self._target_processor.upload_object(
                                obj=data,
                                target_filename=target_key,
                                content_type="text/markdown",
                            )
                        if self.to_formats is None or (
                            self.to_formats and "html" in self.to_formats
                        ):
                            # Export Docling document format to html:
                            target_key = f"html/{name_without_ext}.html"
                            temp_html_file = temp_dir / f"{name_without_ext}.html"

                            conv_res.document.save_as_html(temp_html_file)
                            self._target_processor.upload_file(
                                filename=temp_html_file,
                                target_filename=target_key,
                                content_type="text/html",
                            )

                        if self.to_formats is None or (
                            self.to_formats and "text" in self.to_formats
                        ):
                            # Export Docling document format to text:
                            target_key = f"txt/{name_without_ext}.txt"

                            data = conv_res.document.export_to_text()
                            self._target_processor.upload_object(
                                obj=data,
                                target_filename=target_key,
                                content_type="text/plain",
                            )
                        if self.export_parquet_file:
                            logging.info("saving document info in dataframe...")
                            # Save Docling parquet info into DataFrame:
                            pd_d = self.document_to_dataframe(
                                conv_res=conv_res,
                                pd_dataframe=pd_d,
                                filename=name_without_ext,
                            )

                        yield f"{doc_hash} - SUCCESS"

                    elif conv_res.status == ConversionStatus.PARTIAL_SUCCESS:
                        yield f"{conv_res.input.file} - PARTIAL_SUCCESS"
                    else:
                        yield f"{conv_res.input.file} - FAILURE"

        finally:
            if self.export_parquet_file and not pd_d.empty:
                self.upload_parquet_file(pd_d)

    def upload_page_images(
        self,
        pages: dict[int, PageItem],
        doc_hash: str,
    ):
        for page_no, page in pages.items():
            try:
                if page.image and page.image.pil_image:
                    page_hash = create_hash(f"{doc_hash}_page_no_{page_no}")
                    page_dpi = page.image.dpi
                    page_path_suffix = f"pages/{page_hash}_{page_dpi}.png"
                    buf = BytesIO()
                    page.image.pil_image.save(buf, format="PNG")
                    buf.seek(0)
                    self._target_processor.upload_object(
                        obj=buf,
                        target_filename=page_path_suffix,
                        content_type="application/png",
                    )
                    page.image.uri = Path(".." + page_path_suffix)

            except Exception as exc:
                logging.error(
                    "Upload image of page with hash %r raised error: %r",
                    page_hash,
                    exc,
                )

    def upload_pictures(
        self,
        document: DoclingDocument,
        doc_hash: str,
    ):
        picture_number = 0
        for element, _level in document.iterate_items():
            if isinstance(element, PictureItem):
                if element.image and element.image.pil_image:
                    try:
                        element_hash = create_hash(f"{doc_hash}_img_{picture_number}")
                        element_dpi = element.image.dpi
                        element_path_suffix = f"images/{element_hash}_{element_dpi}.png"
                        buf = BytesIO()
                        element.image.pil_image.save(buf, format="PNG")
                        buf.seek(0)
                        self._target_processor.upload_object(
                            obj=buf,
                            target_filename=element_path_suffix,
                            content_type="application/png",
                        )
                        element.image.uri = Path(".." + element_path_suffix)

                    except Exception as exc:
                        logging.error(
                            "Upload picture with hash %r raised error: %r",
                            element_hash,
                            exc,
                        )
                    picture_number += 1

    def document_to_dataframe(
        self, conv_res: ConversionResult, pd_dataframe: DataFrame, filename: str
    ) -> DataFrame:
        result_table: list[dict[str, Any]] = []

        page_images = []
        for page_no, page in conv_res.document.pages.items():
            if page.image is not None and page.image.pil_image is not None:
                page_images.append(page.image.pil_image.tobytes())

        # Count the number of picture of each type
        num_formulas = 0
        num_codes = 0
        picture_classes = dict.fromkeys(classifier_labels, 0)
        for element, _level in conv_res.document.iterate_items():
            if isinstance(element, PictureItem):
                element.image = None  # reset images
                classification = next(
                    (
                        annot
                        for annot in element.annotations
                        if isinstance(annot, PictureClassificationData)
                    ),
                    None,
                )
                if classification is None or len(classification.predicted_classes) == 0:
                    continue

                predicted_class = classification.predicted_classes[0].class_name
                if predicted_class in picture_classes:
                    picture_classes[predicted_class] += 1

            elif isinstance(element, DocItem):
                if element.label == DocItemLabel.FORMULA:
                    num_formulas += 1
                elif element.label == DocItemLabel.CODE:
                    num_codes += 1

        num_pages = len(conv_res.document.pages)
        num_tables = len(conv_res.document.tables)
        num_elements = len(conv_res.document.texts)
        num_pictures = len(conv_res.document.pictures)

        # All features
        features = [
            num_pages,
            num_elements,
            num_tables,
            num_pictures,
            num_formulas,
            num_codes,
            *picture_classes.values(),
        ]

        doc_hash = (
            conv_res.document.origin.binary_hash
            if conv_res.document.origin
            else "unknown_hash"
        )
        doc_json = json.dumps(conv_res.document.export_to_dict())

        pdf_byte_array: bytearray | None = None
        if os.path.exists(conv_res.input.file):
            with open(conv_res.input.file, "rb") as file:
                pdf_byte_array = bytearray(file.read())

        result_table.append(
            {
                "filename": filename,
                "pdf": pdf_byte_array,
                "doc_hash": doc_hash,
                "document": doc_json,
                "page_images": page_images,
                "features": features,
                "doctags": str.encode(conv_res.document.export_to_document_tokens()),
            }
        )

        pd_df = pd.json_normalize(result_table)
        pd_df = pd_dataframe._append(pd_df)

        return pd_df

    def upload_parquet_file(self, pd_dataframe: DataFrame):
        # Variables to track the file writing process
        file_index = 0
        current_file_size = 0
        current_df = pd.DataFrame()
        # Manifest dictionary
        manifest = {}
        # Current time
        timestamp = datetime.now().strftime("%Y-%m-%dT%H:%M:%SZ")

        while len(pd_dataframe) > 0:
            # Get a chunk of the DataFrame that fits within the file size limit
            chunk_size = min(
                len(pd_dataframe), MAX_PARQUET_FILE_SIZE // (current_file_size + 1)
            )

            # If the chunk size is 0, it means the current file size has exceeded the limit
            if chunk_size == 0:
                with tempfile.NamedTemporaryFile(
                    suffix=f".parquet_{file_index}", dir=self.scratch_dir
                ) as temp_file:
                    pd_dataframe.to_parquet(temp_file)
                    current_file_size = temp_file.seek(0, 2)
                    file_index += 1

                    parquet_file_name = f"{timestamp}_{file_index}.parquet"
                    target_key = f"parquet/{parquet_file_name}"
                    self._target_processor.upload_file(
                        filename=temp_file.name,
                        target_filename=target_key,
                        content_type="application/vnd.apache.parquet",
                    )

                    manifest[f"{parquet_file_name}"] = {
                        "filename": pd_dataframe["filename"].tolist(),
                        "doc_hash": pd_dataframe["doc_hash"].tolist(),
                        "row_number": 3,
                        "timestamp": timestamp,
                    }

                pd_dataframe = pd.DataFrame()
            else:
                # Get the current chunk of the DataFrame
                current_df = pd_dataframe.iloc[:chunk_size]
                pd_dataframe = pd_dataframe.iloc[chunk_size:]

                with tempfile.NamedTemporaryFile(
                    suffix=f".parquet_{file_index}", dir=self.scratch_dir
                ) as temp_file:
                    current_df.to_parquet(temp_file.name)
                    current_file_size = temp_file.seek(0, 2)
                    file_index += 1

                    parquet_file_name = f"{timestamp}_{file_index}.parquet"
                    target_key = f"parquet/{parquet_file_name}"
                    self._target_processor.upload_file(
                        filename=temp_file.name,
                        target_filename=target_key,
                        content_type="application/vnd.apache.parquet",
                    )

                    manifest[f"{parquet_file_name}"] = {
                        "filenames": current_df["filename"].tolist(),
                        "doc_hashes": current_df["doc_hash"].tolist(),
                        "row_number": 3,
                        "timestamp": timestamp,
                    }

        logging.info(f"Total parquet files uploaded: {file_index}")

        # Export manifest file:
        with tempfile.NamedTemporaryFile(
            suffix=".json", dir=self.scratch_dir
        ) as temp_file_json:
            with open(temp_file_json.name, "w") as file:
                json.dump(manifest, file, indent=4)
            self._target_processor.upload_file(
                filename=temp_file_json.name,
                target_filename=f"manifest/{timestamp}.json",
                content_type="application/json",
            )



import asyncio
import base64
import logging
import uuid
import warnings
from pathlib import Path
from typing import Optional

import msgpack
import redis
import redis.asyncio as async_redis
from pydantic import BaseModel
from rq import Queue
from rq.job import Job, JobStatus

from docling.datamodel.base_models import DocumentStream

from docling_jobkit.datamodel.chunking import BaseChunkerOptions, ChunkingExportOptions
from docling_jobkit.datamodel.convert import ConvertDocumentsOptions
from docling_jobkit.datamodel.http_inputs import FileSource, HttpSource
from docling_jobkit.datamodel.result import DoclingTaskResult
from docling_jobkit.datamodel.task import Task, TaskSource, TaskTarget
from docling_jobkit.datamodel.task_meta import TaskStatus, TaskType
from docling_jobkit.orchestrators.base_orchestrator import (
    BaseOrchestrator,
    TaskNotFoundError,
)

_log = logging.getLogger(__name__)


class RQOrchestratorConfig(BaseModel):
    redis_url: str = "redis://localhost:6379/"
    results_ttl: int = 3_600 * 4
    results_prefix: str = "docling:results"
    sub_channel: str = "docling:updates"
    scratch_dir: Optional[Path] = None


class _TaskUpdate(BaseModel):
    task_id: str
    task_status: TaskStatus
    result_key: Optional[str] = None


class RQOrchestrator(BaseOrchestrator):
    @staticmethod
    def make_rq_queue(config: RQOrchestratorConfig) -> tuple[redis.Redis, Queue]:
        conn = redis.from_url(config.redis_url)
        rq_queue = Queue(
            "convert",
            connection=conn,
            default_timeout=14400,
            result_ttl=config.results_ttl,
        )
        return conn, rq_queue

    def __init__(
        self,
        config: RQOrchestratorConfig,
    ):
        super().__init__()
        self.config = config
        self._redis_conn, self._rq_queue = self.make_rq_queue(self.config)
        self._async_redis_conn = async_redis.from_url(self.config.redis_url)
        self._task_result_keys: dict[str, str] = {}

    async def notify_end_job(self, task_id):
        # TODO: check if this is necessary
        pass

    async def enqueue(
        self,
        sources: list[TaskSource],
        target: TaskTarget,
        task_type: TaskType = TaskType.CONVERT,
        options: ConvertDocumentsOptions | None = None,
        convert_options: ConvertDocumentsOptions | None = None,
        chunking_options: BaseChunkerOptions | None = None,
        chunking_export_options: ChunkingExportOptions | None = None,
    ) -> Task:
        if options is not None and convert_options is None:
            convert_options = options
            warnings.warn(
                "'options' is deprecated and will be removed in a future version. "
                "Use 'conversion_options' instead.",
                DeprecationWarning,
                stacklevel=2,
            )
        task_id = str(uuid.uuid4())
        rq_sources: list[HttpSource | FileSource] = []
        for source in sources:
            if isinstance(source, DocumentStream):
                encoded_doc = base64.b64encode(source.stream.read()).decode()
                rq_sources.append(
                    FileSource(filename=source.name, base64_string=encoded_doc)
                )
            elif isinstance(source, (HttpSource | FileSource)):
                rq_sources.append(source)
        chunking_export_options = chunking_export_options or ChunkingExportOptions()
        task = Task(
            task_id=task_id,
            task_type=task_type,
            sources=rq_sources,
            convert_options=convert_options,
            chunking_options=chunking_options,
            chunking_export_options=chunking_export_options,
            target=target,
        )
        self.tasks.update({task.task_id: task})
        task_data = task.model_dump(mode="json", serialize_as_any=True)
        self._rq_queue.enqueue(
            "docling_jobkit.orchestrators.rq.worker.docling_task",
            kwargs={"task_data": task_data},
            job_id=task_id,
            timeout=14400,
        )
        await self.init_task_tracking(task)

        return task

    async def queue_size(self) -> int:
        return self._rq_queue.count

    async def _update_task_from_rq(self, task_id: str) -> None:
        task = await self.get_raw_task(task_id=task_id)
        if task.is_completed():
            return

        job = Job.fetch(task_id, connection=self._redis_conn)
        status = job.get_status()

        if status == JobStatus.FINISHED:
            result = job.latest_result()
            if result is not None and result.type == result.Type.SUCCESSFUL:
                task.set_status(TaskStatus.SUCCESS)
                task_result_key = str(result.return_value)
                self._task_result_keys[task_id] = task_result_key
            else:
                task.set_status(TaskStatus.FAILURE)

        elif status in (
            JobStatus.QUEUED,
            JobStatus.SCHEDULED,
            JobStatus.STOPPED,
            JobStatus.DEFERRED,
        ):
            task.set_status(TaskStatus.PENDING)
        elif status == JobStatus.STARTED:
            task.set_status(TaskStatus.STARTED)
        else:
            task.set_status(TaskStatus.FAILURE)

    async def task_status(self, task_id: str, wait: float = 0.0) -> Task:
        await self._update_task_from_rq(task_id=task_id)
        return await self.get_raw_task(task_id=task_id)

    async def get_queue_position(self, task_id: str) -> Optional[int]:
        try:
            job = Job.fetch(task_id, connection=self._redis_conn)
            queue_pos = job.get_position()
            return queue_pos + 1 if queue_pos is not None else None
        except Exception as e:
            _log.error("An error occour getting queue position.", exc_info=e)
            return None

    async def task_result(
        self,
        task_id: str,
    ) -> Optional[DoclingTaskResult]:
        if task_id not in self._task_result_keys:
            return None
        result_key = self._task_result_keys[task_id]
        packed = await self._async_redis_conn.get(result_key)
        result = DoclingTaskResult.model_validate(
            msgpack.unpackb(packed, raw=False, strict_map_key=False)
        )
        return result

    async def _listen_for_updates(self):
        pubsub = self._async_redis_conn.pubsub()

        # Subscribe to a single channel
        await pubsub.subscribe(self.config.sub_channel)

        _log.debug("Listening for updates...")

        # Listen for messages
        async for message in pubsub.listen():
            if message["type"] == "message":
                data = _TaskUpdate.model_validate_json(message["data"])
                try:
                    task = await self.get_raw_task(task_id=data.task_id)
                    if task.is_completed():
                        _log.debug("Task already completed. No update will be done.")
                        continue

                    # Update the status
                    task.set_status(data.task_status)
                    # Update the results lookup
                    if (
                        data.task_status == TaskStatus.SUCCESS
                        and data.result_key is not None
                    ):
                        self._task_result_keys[data.task_id] = data.result_key

                    if self.notifier:
                        # Notify clients about task updates
                        await self.notifier.notify_task_subscribers(task.task_id)

                        # Notify clients about queue updates
                        await self.notifier.notify_queue_positions()

                except TaskNotFoundError:
                    _log.warning(f"Task {data.task_id} not found.")

    async def process_queue(self):
        # Create a pool of workers
        pubsub_worker = []
        _log.debug("PubSub worker starting.")
        pubsub_worker = asyncio.create_task(self._listen_for_updates())

        # Wait for all worker to complete
        await asyncio.gather(pubsub_worker)
        _log.debug("PubSub worker completed.")

    async def delete_task(self, task_id: str):
        _log.info(f"Deleting result of task {task_id=}")

        # Delete the result data from Redis if it exists
        if task_id in self._task_result_keys:
            await self._async_redis_conn.delete(self._task_result_keys[task_id])
            del self._task_result_keys[task_id]

        # Delete the RQ job itself to free up Redis memory
        # This includes the job metadata and result stream
        try:
            job = Job.fetch(task_id, connection=self._redis_conn)
            job.delete()
            _log.debug(f"Deleted RQ job {task_id=}")
        except Exception as e:
            # Job may not exist or already be deleted - this is not an error
            _log.debug(f"Could not delete RQ job {task_id=}: {e}")

        await super().delete_task(task_id)

    async def warm_up_caches(self):
        pass

    async def check_connection(self):
        # Check redis connection is up
        try:
            self._redis_conn.ping()
        except Exception:
            raise RuntimeError("No connection to Redis")

    async def clear_converters(self):
        pass



from pathlib import Path
from typing import Iterator, TypedDict

from docling_core.types.io import DocumentStream

from docling_jobkit.connectors.source_processor import BaseSourceProcessor
from docling_jobkit.datamodel.task_sources import TaskLocalPathSource


def _should_ignore_file(file_path: Path) -> bool:
    """
    Check if a file should be ignored based on common patterns for
    hidden files, temporary files, and system metadata files.

    Returns True if the file should be ignored, False otherwise.
    """
    name = file_path.name

    # Hidden files (starting with .)
    if name.startswith("."):
        return True

    # Vim temporary files
    if name.endswith(("~", ".swp", ".swo")):
        return True

    # Emacs temporary files
    if name.startswith("#") and name.endswith("#"):
        return True

    # Microsoft Office temporary files
    if name.startswith("~$"):
        return True

    # Windows thumbnail cache
    if name.lower() == "thumbs.db":
        return True

    # Desktop.ini (Windows)
    if name.lower() == "desktop.ini":
        return True

    return False


class LocalPathFileIdentifier(TypedDict):
    path: Path
    size: int
    last_modified: float


class LocalPathSourceProcessor(BaseSourceProcessor[LocalPathFileIdentifier]):
    def __init__(self, source: TaskLocalPathSource):
        super().__init__()
        self._source = source

    def _initialize(self):
        """Validate that the path exists."""
        if not self._source.path.exists():
            raise FileNotFoundError(f"Path does not exist: {self._source.path}")

    def _finalize(self):
        """No cleanup needed for local filesystem."""

    def _list_document_ids(self) -> Iterator[LocalPathFileIdentifier]:
        """
        List all files based on the source configuration.
        - If path is a file, yield that single file
        - If path is a directory, discover files based on pattern and recursive settings
        """
        path = self._source.path

        if path.is_file():
            # Single file case
            stat = path.stat()
            yield LocalPathFileIdentifier(
                path=path,
                size=stat.st_size,
                last_modified=stat.st_mtime,
            )
        elif path.is_dir():
            # Directory case - use glob or rglob based on recursive setting
            if self._source.recursive:
                # Recursive traversal
                files = path.rglob(self._source.pattern)
            else:
                # Non-recursive traversal
                files = path.glob(self._source.pattern)

            for file_path in files:
                # Only yield actual files, not directories
                # Skip hidden files, temporary files, and system metadata
                if file_path.is_file() and not _should_ignore_file(file_path):
                    stat = file_path.stat()
                    yield LocalPathFileIdentifier(
                        path=file_path,
                        size=stat.st_size,
                        last_modified=stat.st_mtime,
                    )
        else:
            raise ValueError(f"Path is neither a file nor a directory: {path}")

    def _count_documents(self) -> int:
        """Count total number of documents."""
        return sum(1 for _ in self._list_document_ids())

    def _fetch_document_by_id(
        self, identifier: LocalPathFileIdentifier
    ) -> DocumentStream:
        """Fetch a document by opening the file from the local filesystem."""
        file_path = identifier["path"]

        # Open file in binary mode and return as DocumentStream
        with open(file_path, "rb") as f:
            content = f.read()

        from io import BytesIO

        buffer = BytesIO(content)

        return DocumentStream(name=str(file_path), stream=buffer)

    def _fetch_documents(self) -> Iterator[DocumentStream]:
        """Iterate through all documents."""
        for identifier in self._list_document_ids():
            yield self._fetch_document_by_id(identifier)



import hashlib
import logging
import threading
import time
from functools import lru_cache
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional

from pydantic import BaseModel, Field

from docling.datamodel.base_models import ConversionStatus
from docling.datamodel.document import ConversionResult
from docling_core.transforms.chunker import BaseChunker
from docling_core.transforms.chunker.hierarchical_chunker import (
    ChunkingSerializerProvider,
    DocChunk,
    HierarchicalChunker,
)
from docling_core.transforms.chunker.hybrid_chunker import HybridChunker
from docling_core.transforms.chunker.tokenizer.huggingface import (
    HuggingFaceTokenizer,
)
from docling_core.types.doc.document import DoclingDocument, ImageRefMode

from docling_jobkit.convert.results import _export_document_as_content
from docling_jobkit.datamodel.chunking import (
    BaseChunkerOptions,
    HierarchicalChunkerOptions,
    HybridChunkerOptions,
)
from docling_jobkit.datamodel.convert import ConvertDocumentsOptions
from docling_jobkit.datamodel.result import (
    ChunkedDocumentResult,
    ChunkedDocumentResultItem,
    DoclingTaskResult,
    ExportDocumentResponse,
    ExportResult,
)
from docling_jobkit.datamodel.task import Task
from docling_jobkit.datamodel.task_targets import InBodyTarget

_log = logging.getLogger(__name__)


class MarkdownTableSerializerProvider(ChunkingSerializerProvider):
    """Serializer provider that uses markdown table format for table serialization."""

    def get_serializer(self, doc):
        """Get a serializer that uses markdown table format."""
        from docling_core.transforms.chunker.hierarchical_chunker import (
            ChunkingDocSerializer,
        )
        from docling_core.transforms.serializer.markdown import MarkdownTableSerializer

        return ChunkingDocSerializer(
            doc=doc,
            table_serializer=MarkdownTableSerializer(),
        )


class DocumentChunkerConfig(BaseModel):
    """Configuration for DocumentChunker."""

    cache_size: int = Field(
        default=10,
        gt=0,
        le=100,
        description="Maximum number of chunker instances to cache",
    )


class DocumentChunkerManager:
    """Handles document chunking for RAG workflows using chunkers from docling-core."""

    def __init__(self, config: Optional[DocumentChunkerConfig] = None):
        self.config = config or DocumentChunkerConfig()
        self._cache_lock = threading.Lock()
        self._options_map: dict[bytes, BaseChunkerOptions] = {}
        self._get_chunker_from_cache = self._create_chunker_cache()

    def _create_chunker_cache(self):
        """Create LRU cache for chunker instances."""

        @lru_cache(maxsize=self.config.cache_size)
        def _get_chunker_from_cache(cache_key: bytes) -> BaseChunker:
            try:
                options = self._options_map[cache_key]

                # Create serializer provider based on markdown table option
                if options.use_markdown_tables:
                    serializer_provider: ChunkingSerializerProvider = (
                        MarkdownTableSerializerProvider()
                    )
                else:
                    serializer_provider = ChunkingSerializerProvider()

                if isinstance(options, HybridChunkerOptions):
                    # Create tokenizer
                    tokenizer_name = options.tokenizer
                    tokenizer_obj = HuggingFaceTokenizer.from_pretrained(
                        model_name=tokenizer_name,
                        max_tokens=options.max_tokens,
                    )

                    chunker: BaseChunker = HybridChunker(
                        tokenizer=tokenizer_obj,
                        merge_peers=options.merge_peers,
                        serializer_provider=serializer_provider,
                    )
                elif isinstance(options, HierarchicalChunkerOptions):
                    chunker = HierarchicalChunker(
                        serializer_provider=serializer_provider
                    )
                else:
                    raise RuntimeError(f"Unknown chunker {options.chunker}.")

                return chunker

            except ImportError as e:
                _log.error(f"Missing dependencies for document chunking: {e}")
                raise ImportError(
                    "Document chunking requires docling-core with chunking dependencies. "
                    "Install with: pip install 'docling-core[chunking]'"
                ) from e
            except (ValueError, TypeError, AttributeError) as e:
                _log.error(f"Invalid chunking configuration: {e}")
                raise ValueError(f"Invalid chunking options: {e}") from e
            except (OSError, ConnectionError) as e:
                _log.error(f"Resource or connection error during chunker creation: {e}")
                raise RuntimeError(
                    f"Failed to initialize chunker resources: {e}"
                ) from e

        return _get_chunker_from_cache

    def _get_chunker(self, options: BaseChunkerOptions) -> BaseChunker:
        """Get or create a cached BaseChunker instance."""
        # Create a cache key based on chunking options using the same pattern as the repo
        cache_key = self._generate_cache_key(options)

        with self._cache_lock:
            self._options_map[cache_key] = options
            return self._get_chunker_from_cache(cache_key)

    def _generate_cache_key(self, options: BaseChunkerOptions) -> bytes:
        """Generate a deterministic cache key from chunking options."""
        serialized_data = options.model_dump_json(serialize_as_any=True)
        options_hash = hashlib.sha1(
            serialized_data.encode(), usedforsecurity=False
        ).digest()
        return options_hash

    def clear_cache(self):
        """Clear the chunker cache."""
        with self._cache_lock:
            self._get_chunker_from_cache.cache_clear()

    def chunk_document(
        self,
        document: DoclingDocument,
        filename: str,
        options: BaseChunkerOptions,
    ) -> Iterable[ChunkedDocumentResultItem]:
        """Chunk a document using chunker from docling-core."""

        chunker = self._get_chunker(options)

        chunks = list(chunker.chunk(document))

        # Convert chunks to response format
        chunk_items: list[ChunkedDocumentResultItem] = []
        for i, chunk in enumerate(chunks):
            page_numbers: List[int] = []
            metadata: Dict[str, Any] = {}

            doc_chunk = DocChunk.model_validate(chunk)

            # Extract page numbers and doc_items refs
            page_numbers = []
            doc_items = []
            for item in doc_chunk.meta.doc_items:
                doc_items.append(item.self_ref)
                for prov in item.prov:
                    page_numbers.append(prov.page_no)

            # Remove duplicates and sort
            page_numbers = sorted(set(page_numbers))

            # Store additional metadata
            if doc_chunk.meta.origin:
                metadata["origin"] = doc_chunk.meta.origin

            # Get the text
            text = chunker.contextualize(doc_chunk)

            # Compute the number of tokens
            num_tokens: int | None = None
            if isinstance(chunker, HybridChunker):
                num_tokens = chunker.tokenizer.count_tokens(text)

            # Create chunk item
            chunk_item = ChunkedDocumentResultItem(
                filename=filename,
                chunk_index=i,
                text=text,
                raw_text=doc_chunk.text if options.include_raw_text else None,
                num_tokens=num_tokens,
                headings=doc_chunk.meta.headings,
                captions=doc_chunk.meta.captions,
                doc_items=doc_items,
                page_numbers=page_numbers,
                metadata=metadata,
            )
            chunk_items.append(chunk_item)

        return chunk_items


def process_chunk_results(
    task: Task,
    conv_results: Iterable[ConversionResult],
    work_dir: Path,
    chunker_manager: Optional[DocumentChunkerManager] = None,
) -> DoclingTaskResult:
    # Let's start by processing the documents
    start_time = time.monotonic()
    chunking_options = task.chunking_options or HybridChunkerOptions()
    conversion_options = task.convert_options or ConvertDocumentsOptions()

    # We have some results, let's prepare the response
    task_result: ChunkedDocumentResult
    chunks: list[ChunkedDocumentResultItem] = []
    documents: list[ExportResult] = []
    num_succeeded = 0
    num_failed = 0

    # TODO: DocumentChunkerManager should be initialized outside for really working as a cache
    chunker_manager = chunker_manager or DocumentChunkerManager()
    for conv_res in conv_results:
        errors = conv_res.errors
        filename = conv_res.input.file.name
        if conv_res.status == ConversionStatus.SUCCESS:
            try:
                chunks.extend(
                    chunker_manager.chunk_document(
                        document=conv_res.document,
                        filename=filename,
                        options=chunking_options,
                    )
                )
                num_succeeded += 1
            except Exception as e:
                _log.exception(
                    f"Document chunking failed for {conv_res.input.file}: {e}",
                    stack_info=True,
                )
                num_failed += 1
                # TODO: for propagating errors we have first to allow other component_type in the Docling class.
                # errors = [
                #     *errors,
                #     ErrorItem(
                #         component_type="chunking",
                #         module_name=type(e).__name__,
                #         error_message=str(e),
                #     ),
                # ]

        else:
            _log.warning(f"Document {conv_res.input.file} failed to convert.")
            num_failed += 1

        if (
            isinstance(task.target, InBodyTarget)
            and task.chunking_export_options.include_converted_doc
        ):
            if conversion_options.image_export_mode == ImageRefMode.REFERENCED:
                raise RuntimeError("InBodyTarget cannot use REFERENCED image mode.")

            doc_content = _export_document_as_content(
                conv_res,
                export_json=True,
                export_doctags=False,
                export_html=False,
                export_md=False,
                export_txt=False,
                image_mode=conversion_options.image_export_mode,
                md_page_break_placeholder=conversion_options.md_page_break_placeholder,
            )
        else:
            doc_content = ExportDocumentResponse(filename=filename)

        doc_result = ExportResult(
            content=doc_content,
            status=conv_res.status,
            timings=conv_res.timings,
            errors=errors,
        )

        documents.append(doc_result)

    num_total = num_succeeded + num_failed
    processing_time = time.monotonic() - start_time
    _log.info(
        f"Processed {num_total} docs generating {len(chunks)} chunks in {processing_time:.2f} seconds."
    )

    if isinstance(task.target, InBodyTarget):
        task_result = ChunkedDocumentResult(
            chunks=chunks,
            documents=documents,
            processing_time=processing_time,
            chunking_info=chunking_options.model_dump(mode="json"),
        )

    # Multiple documents were processed, or we are forced returning as a file
    else:
        raise NotImplementedError("Saving chunks to a file is not yet supported.")

    return DoclingTaskResult(
        result=task_result,
        processing_time=processing_time,
        num_succeeded=num_succeeded,
        num_failed=num_failed,
        num_converted=num_total,
    )



from io import BytesIO
from pathlib import Path
from typing import BinaryIO

from docling_jobkit.connectors.s3_helper import get_s3_connection
from docling_jobkit.connectors.target_processor import BaseTargetProcessor
from docling_jobkit.datamodel.s3_coords import S3Coordinates


class S3TargetProcessor(BaseTargetProcessor):
    def __init__(self, coords: S3Coordinates):
        super().__init__()
        self._coords = coords

    def _initialize(self):
        self._client, self._resource = get_s3_connection(self._coords)

    def _finalize(self):
        self._client.close()

    def upload_file(
        self,
        filename: str | Path,
        target_filename: str,
        content_type: str,
    ) -> None:
        """
        Upload a local file from disk into the S3 bucket.
        """
        full_key = (
            f"{self._coords.key_prefix}{target_filename}"
            if self._coords.key_prefix
            else target_filename
        )
        self._client.upload_file(
            Filename=filename,
            Bucket=self._coords.bucket,
            Key=full_key,
            ExtraArgs={"ContentType": content_type},
        )

    def upload_object(
        self,
        obj: str | bytes | BinaryIO,
        target_filename: str,
        content_type: str,
    ) -> None:
        """
        Upload an in-memory object (bytes or file-like) into the S3 bucket.
        """
        full_key = (
            f"{self._coords.key_prefix}{target_filename}"
            if self._coords.key_prefix
            else target_filename
        )
        if isinstance(obj, (bytes, bytearray)):
            body: BinaryIO = BytesIO(obj)
        elif isinstance(obj, str):
            body = BytesIO(obj.encode())
        else:
            body = obj  # assume it's a file-like object

        self._client.upload_fileobj(
            Fileobj=body,
            Bucket=self._coords.bucket,
            Key=full_key,
            ExtraArgs={"ContentType": content_type},
        )
